/// GWAV intefrace class definition
/**
 * @file 
 * This file defines the interface class for the gwav format classes
 *
 * @author Gabor Ligeti
 *
 */
 
// Can't use standard include guard if it is to be used in two versions  
#pragma once
//#ifndef _IGWAV_H_
//#define _IGWAV_H_

#include "cmgwstatus.h"
#include <stdint.h>
#include <stdlib.h>

#include "gwavns.h"
namespace cmg{ namespace GWAV_NS {  

/**
 * @brief Wavelet encoder interface
 *
 * @details An untemplated interface for Wavelet with DWT transformation and compression
 */
class IGWavCompressor
{
public:

  virtual ~IGWavCompressor() {};
/**
 * @brief Starts performing the compression
 * 
 * @details Starts the compression. It can be stopped any time (from another thread) by calling stop()
 *          which will suspend it (will return with WaveletStatus_Suspended, calling compress()
 *          again will continue
 *
 * @param segments  number of segments to process
 *
 * @returns success status.
 */
  virtual WaveletStatus compress(int segments=0) = 0;

/// Request to stop (suspend) processing. Calling the encoder/decoder again will continue.
  virtual void stop() = 0; 


/// Request to stop/kill processing. Can not continue after a kill() call
  virtual void kill() = 0;

public:


/**
 * @brief   Instance creator
 *
 * @details Creates an IGWavCompressor instance
 *
 * @param bytes        bytes per pixel (as bypp in the header0
 * @param bits         bits per pixel (as bipp in the header)
 * @param xsize        x size
 * @param ysize        y size
 * @param imgdata      source raw image data
 * @param bitdatasize  target buffer size. If 0, do lossless.
 * @param dodelete     if bitdata is null, should the destructor delete the allocated buffer
 *
 * @return             an IGWavCompressor object instance pointer
 */
  static IGWavCompressor *createCompressor( int xsize, int ysize, int bits, int bytes, const void * imgdata, size_t bitdatasize = 0, bool dodelete = true); 
  
  
/**
 * @brief Releases an IGwavCompressor instance
 * 
 * @param compressor  IGwavCompressor object instance pointer
 */
  static void release( IGWavCompressor *compressor);
};



/**
 * @brief Wavelet decoder interface
 *
 * @details An untemplated interface for Wavelet with DWT transformation and decompression
 */
class IGWavDecompressor
{
public:

/**
 * @name Member access functions
 * Member access functions to get the image properties
 */
///@{

/**
 * @brief   Get bytes per pixel
 *
 * @returns        bytes per pixel
 */
  virtual int getBytesPerPixel() = 0;
  
/**
 * @brief   Get bits per pixel
 *
 * @returns        bits per pixel
 */
  virtual int getBitsPerPixel() = 0;
  
/**
 * @brief   Get x size of image
 *
 * @returns        x size
 */
  virtual int getXSize() = 0;

/**
 * @brief   Get y size of image
 *
 * @returns        y size
 */
  virtual int getYSize() = 0;

///@} // Member access functions
 

public:

  virtual ~IGWavDecompressor(){};

/**
 * @brief Starts performing the decompression
 * 
 * @details Starts the decompression. It can be stopped any time (from another thread) by calling stop()
 *          which will suspend it (will return with WaveletStatus_Suspended, calling decompress()
 *          again will continue
 *
 * @param imgdata     target raw image data buffer
 * @param imgdatasize target buffer size. Must match xsize*ysize*bytesperpixel
 * @param segments  number of segments to process
 *
 *
 * @returns           success status.
 */
  virtual WaveletStatus decompress( void *imgdata, size_t imgdatasize, int segments=0) = 0;
  
/// Request to stop (suspend) processing. Calling the encoder/decoder again will continue.
  virtual void stop() = 0; 

/// Request to stop/kill processing. Can not continue after a kill() call
  virtual void kill() = 0;

public:


/**
 * @brief   Instance creator
 *
 * @details Creates an IGWavDecompressor instance
 * 
 * @param bitdata     pointer to the compressed stream buffer
 * @param bitdatasize size of the available stream data in the buffer
 * @param factor      decompression scale factor exponent (0->1:1, 1->1:2, 2->1:4, etc)
 * @param rfactor     resample portion of decompression scale
 *
 * @returns           an IGWavDecompressor object instance pointer
 */
  static IGWavDecompressor *createDecompressor( const uint8_t *bitdata, size_t bitdatasize, int factor = 0, int rfactor = 0);


/**
 * @brief Releases an IGwavDecompressor instance
 * 
 * @param decompressor  IGwavDecompressor object instance pointer
 */
  static void release( IGWavDecompressor *decompressor);

/**
 * @brief gets the header size (minimum needed from a stream)
 * 
 * @return                 the size of the header
 */
  static size_t getHeaderSize();

/**
 * @brief gets info from a compressed bitstream (header)
 * 
 * @param bitstream        pointer to the bitstream
 * @param compresslength   size of the (available) bistream
 * @param xsize            pointer to store the image X size in (unless NULL)
 * @param ysize            pointer to store the image Y size in (unless NULL)
 * @param outputsize       pointer to store the required output size in (unless NULL)
 * @param bytesperpixel    pointer to store the image bytes per pixel (unless NULL)
 * @param bitsperpixel     pointer to store the image bits per pixel (unless NULL)
 * @param factor           decompression scale factor exponent (0->1:1, 1->1:2, 2->1:4, etc)
 *
 * @return                 true if the stream is valid
 */
  static bool getInfo( const uint8_t *bitstream, size_t compresslength, int *xsize = NULL, int *ysize = NULL, size_t *outputsize = NULL, int * bytesperpixel = NULL, int * bitsperpixel = NULL, int factor = 0);
  
/**
 * @brief Validates a compressed bitstream (legacy, deprecated)
 * 
 * @param bitstream        pointer to the bitstream
 * @param compresslength   size of the (available) bistream
 * @param xsize            pointer to store the image X size in (unless NULL)
 * @param ysize            pointer to store the image Y size in (unless NULL)
 * @param outputsize       pointer to store the required output size in (unless NULL)
 *
 * @return                 true if the stream is valid
 */
 
  static bool validate( const unsigned char *bitstream, unsigned long compresslength, int *xsize = NULL, int *ysize = NULL, size_t *outputsize = NULL) 
  {
    return getInfo( (const uint8_t *)bitstream, (size_t)compresslength, xsize, ysize, outputsize);
  }
  
};

} } // namespace cmg::GWAV_NS

//#endif //_IGWAV_H_
