/*
 * Component:  common
 *
 * Purpose:
 *	    Header file for histogram analyzis
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef __HISTOGRAM_H_INCLUDED__
#define __HISTOGRAM_H_INCLUDED__

#define HISTOGRAM_WL_STEP 5
#define HISTOGRAM_WL_NUM 10

#ifndef IMGPROC_NO_DCMTK
#include "dcmtk/ofstd/oftypes.h"
#else
#include "imgproc/nodcmtk.h"
#endif //IMGPROC_NO_DCMTK

#define BIAS 30
#define BASE 1.001

#define BASE_CORR_DEF 1.1
#define EXP_CORR_DEF 1.2
#define RANGE_CORR_DEF 1000
#define BEST_CORR_DEF 0.6

/**
* Applies histogram analyzis on pixel values and calculates Window/Level pairs
*
* HISTOGRAM_WL_NUM (10) Window/Level pair will be calculated started from 100%
* and stepping down by HISTOGRAM_WL_STEP (5) percents.
*
* Calculated Window/Level values can be retrieved by their index value:
*
*		0 - 100%			5 -  75%
*		1 -  95%			6 -  70%
*		2 -  90%			7 -  65%
*		3 -  85%			8 -  60%
*		4 -  80%			9 -  55%
*/
class CHistogram {
private:

	/** Array for holding pixel values */
	Uint32 * m_array;

	/** Number of bytes used by a pixel */
	int m_bytesPerPixel;

	/** Size of pixel value array */
	Uint32 m_arraySize;

	/** Array of calculated Window values */
	Sint32 m_window[HISTOGRAM_WL_NUM];

	/** Array of calculated Level values */
	Sint32 m_level[HISTOGRAM_WL_NUM];

	/** Smallest pixel value in the image */
	Sint32 m_minValue;

	/** Largest pixel value in the image */
	Sint32 m_maxValue;

	/** true if the histogram was created for a signed image */
	bool m_isSigned;

	/** parameters of the aggressive formula */
	double m_dBaseCorr;
	double m_dExpCorr;
	int m_nRealSize;

	/** Array for holding pixel values in aggressive mode */
	Uint32 * m_array2;

	/** Array of calculated Window values in aggressive mode */
	Sint32 m_window2[HISTOGRAM_WL_NUM];

	/** Array of calculated Level values in aggressive mode */
	Sint32 m_level2[HISTOGRAM_WL_NUM];

	/**
	 * Takes the smallest and largest pixel value from the pixel values array and fills.
	 */
	void calculateMinMax();

public:

	/**
	* Constructor for creating a new CHistogram object.
	*/
	CHistogram();

	/**
	* Destructor for destroying a new CHistogram object.
	*/
	~CHistogram();

	/**
	* (Re) initialize CHistogram object.
	* Allocates memory for pixel value array if necessary,
	* or reuse existing, if bytesPerPixel is the same as the
	* current one. Clears calculated Window/Level values.
	*
	* @param bytesPerPixel   Number of bytes used by a pixel
	* @param isSigned        true indicates that the image the
	*                        histogram will be used for contains
	*                        signed pixel values
	*
	* @return true if succeed
	*/
	bool init(int bytesPerPixel, bool isSigned, double dBaseCorr, double dExpCorr, int nRealSize);

	/**
	* Analyze the histogram of the pixel values.
	*
	* @param nPercent   required WL setting contained pixels (0 if all HISTOGRAM_WL_NUM pairs are required)
	*/
	void analyze(int nPercent);

	/**
	* Analyze the histogram of the pixel values.
	*
	* @param pArray     histogram pointer
	* @param nTotal     sum of histogram pointer content
	* @param nPercent   required WL setting contained pixels (0 if all HISTOGRAM_WL_NUM pairs are required)
	* @param pWindow    window array pointer
	* @param pLevel     level array pointer
	*/
	void analyze2(Uint32* pArray, Uint32 nTotal, int nPercent, Sint32* pWindow, Sint32* pLevel);

	/**
	* Increments the counter of the pixel value
	*
	* @param value   Pixel value
	*/
	inline void increment(int value) {
		m_array[value]++;
	}

	/**
	* Increments the counter of the pixel value
	*
	* @param value   Pixel value
	* @param inc     Increment value
	*/
	inline void increment(int value, int inc) {
		m_array[value] += inc;
	}

	/**
	* Retireves the calculated Window value for the given index
	*
	* @param ndx   Index of the calculated Window value to be retrieved (0 if nPercent was nonzero)
	*
	* @return the calculated Window value for the given index
	*/
	inline Sint32 getWindow(int ndx) {
		return m_window[ndx];
	}

	/**
	* Retireves the calculated Level value for the given index
	*
	* @param ndx   Index of the calculated Level value to be retrieved (0 if nPercent was nonzero)
	*
	* @return the calculated Level value for the given index
	*/
	inline Sint32 getLevel(int ndx) {
		return m_level[ndx];
	}

	/**
	* Retireves the aggressive calculated Window value for the given index
	*
	* @param ndx   Index of the calculated Window value to be retrieved (0 if nPercent was nonzero)
	*
	* @return the calculated Window value for the given index
	*/
	inline Sint32 getWindow2(int ndx) {
		return m_window2[ndx];
	}

	/**
	* Retireves the aggressive calculated Level value for the given index
	*
	* @param ndx   Index of the calculated Level value to be retrieved (0 if nPercent was nonzero)
	*
	* @return the calculated Level value for the given index
	*/
	inline Sint32 getLevel2(int ndx) {
		return m_level2[ndx];
	}

	/**
	* Retrieves the number of calculated Window/Level pairs
	*
	* @return HISTOGRAM_WL_NUM (=10)
	*/
	inline int getWindowLevelNum() {
		return HISTOGRAM_WL_NUM;
	}

	/**
	 * Retrieves the smallest pixel value in the image.
	 * The smallest pixel value is only available after
	 * analyze() was called.
	 *
	 * @return the smallest pixel value in the image
	 */
	inline Sint32 getMinValue() {
		return m_minValue;
	}

	/**
	 * Retrieves the largest pixel value in the image
	 * The largest pixel value is only available after
	 * analyze() was called.
	 *
	 * @return the largest pixel value in the image
	 */
	inline Sint32 getMaxValue() {
		return m_maxValue;
	}
};


#endif // __HISTOGRAM_H_INCLUDED__
