#ifndef _ERAD_IMUTILS_CONFIG_H_
#define _ERAD_IMUTILS_CONFIG_H_

#include <vector>
#include <string>
#include <time.h>

#include "IMProperties.h"

namespace erad { namespace imutils {


/**
 * Class to automatically remove properties based configuration files
 * when changed.
 *
 * By default checks the last modification time at every access to the configuration file,
 * but this can be controlled by specifying check interval seconds.
 *
 * Not thread safe.
 */
class Config {
	private:
		/** Path to the configuration file */
		std::vector<std::string> paths;
		/** Last modified time of the configuration file */
		time_t lastModified;
		/** Time of next check */
		time_t nextCheck;
		/** Default check interval (default: 0 seconds) */
		time_t configCheckInterval;

	protected:
		/** Content of the configuration file */
		IMProperties configPrp;

		/** Resets configuration to default value to (re)load the file */
		virtual void reset();
		/** Loads the content of the configuration file */
		virtual void load();
		/** Checks and reloads the configuration file on demand */
		void check();

	public:
		/**
		 * Constructs a new Config with no configuration file specified
		 */
		Config();
		/**
		 * Constructs a new Config with a configuration file specified
		 * @param path   path to the configuration file
		 */
		Config( const std::string& path );
		Config( const std::vector<std::string> & paths );
		virtual ~Config();

		/**
		 * Sets the configuration file
		 * @param path   path to the configuration file
		 */
		inline void setConfigFile( const std::string& path ) {
			this->paths.clear();
			this->paths.push_back( path );
			lastModified = 0;
			nextCheck = 0;
		}

		/**
		 * Sets the configuration file
		 * @param path   path to the configuration file
		 */
		inline void setConfigFiles( const std::vector<std::string> & paths ) {
			this->paths = paths;
			lastModified = 0;
			nextCheck = 0;
		}



		/**
		 * Sets the check interval
		 * @param newValue   new check interval in seconds
		 */
		inline void setConfigCheckInterval( time_t newValue ) {
			configCheckInterval = newValue;
		}

		/**
		 * @return the content of the configuration file
		 */
		inline const IMProperties& getConfigPrp() {
			check();
			return configPrp;
		}

		/**
		 * @return the path to the configuration file
		 */
		inline const std::string& getConfigFile() const {
			return paths.front();
		}
};


} } // namespace

#endif /* _ERAD_IMUTILS_CONFIG_H_ */
