/*
 * Essential utilities for DICOMweb services
 */
#include <string>

#ifndef IMUTILS_DICOMWEB_H_
#define IMUTILS_DICOMWEB_H_

namespace erad::dicomweb {
// Storage failure reason
// Ref: DICOMWeb specification for storage (DICOM PS3.18 2025a - Web Services -> I.2 Store Instances Response Attribute Description)
// https://dicom.nema.org/medical/dicom/2019c/output/chtml/part18/sect_I.2.2.html
typedef enum {
	// Out of Resources <42752 - 43007>
	OUT_OF_RESOURCES,
	// Data Set does not match SOP Class <43264 - 43519>
	SOP_CLASS_MISMATCH,
	// Cannot understand certain Data Elements <49152 - 53247>
	INVALID_DATA_ELEMENT,
	// Referenced Transfer Syntax not supported <49442>
	UNSUPPORTED_TRANSFER_SYNTAX,
	// Processing failure <272>
	PROCESSING_FAILURE,
	// Referenced SOP Class not supported <290>
	UNSUPPORTED_SOP_CLASS,
	// Success
	NO_FAILURE,
	INVALID
} StorageFailureType;

// HTTP response codes. Specific to GCP Healthcare service API
// Ref: https://cloud.google.com/healthcare-api/docs/dicom#status_codes_2
namespace gcphealthcare {

typedef enum {
	// Possibly when DICOMweb returns the failure: DEADLINE_EXCEEDED
	PROCESSING_TIMEOUT,
	// Possibly when DICOMweb returns the failure: already exists
	RESOURCE_ALREADY_EXISTS,
	NO_FAILURE,
	INVALID
} StorageFailureReason;

// Storage request has taken longer time to process than expected
const std::string REQUEST_FAILURE_DEADLINE_EXCEEDED = "DEADLINE_EXCEEDED";

const std::string REQUEST_FAILURE_RESOURCE_EXISTS = "already exists";

/**
 * Returns the type of failure of GCP Healthcare storage service based on the failure code.
 * Ref: https://dicom.nema.org/medical/dicom/2019c/output/chtml/part18/sect_I.2.2.html
 * @param failureCode   Failure code
 * @return              Failure type
 */
StorageFailureType storageFailureCodeToType( const unsigned int& failureCode );

/**
 * Returns the failure reason code mapped to the reason text
 * @param failureReasonText Reason text
 * @return                  ReasonCode
 */
StorageFailureReason storageFailureReasonTextToCode( const std::string& failureReasonText );

/**
 * Parses the DICOMWeb response message (XML) and extracts the failure reason code and detail.
 * Ref: https://dicom.nema.org/medical/dicom/2019c/output/chtml/part18/sect_I.2.2.html
 * @param responseText      Response text
 * @param failureType       Type of failure
 * @param failureReason     Reason text
 * @return                  If the response text is parsed successfully and able to
 *                          extract the reason code and detail.
 */
bool parseResponse( const std::string& responseText,
		StorageFailureType& failureType, StorageFailureReason& failureReason );

}
} // namespace erad::dicomweb

#endif
