/*
 * Interface definition and GCP implementation to access GCP Healthare API/Query
 * custome service.
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_GCPCONNECTOR_H_
#define IMUTILS_INCLUDE_IMUTILS_GCPCONNECTOR_H_

#include "imutils/IMProperties.h"
#include <curl/curl.h>
#include <dcmtk/ofstd/ofstring.h>
#include <google/cloud/credentials.h>
#include <nlohmann/json.hpp>

namespace erad::gcp {
/* ********************************************
 *
 * Interface
 *
 * ********************************************/

/**
 * class used to initialize GCP Connection
 */
class GCPConnector {
	public:
		/* ******************************
		 *
		 * GCPConnector
		 *
		 * ******************************/
		GCPConnector();

		/**
		 * Set curl headers as part of network request initialization.
		 *
		 * @param accessToken   access token
		 * @param contentType   content type
		 *
		 * @return true if success
		 */
		OFBool setCurlHeaders( const OFString& accessToken,
				const OFString& contentType );

		/**
		 * Set up curl as part of network request initialization.
		 *
		 * @param accessToken   access token
		 *
		 * @return true if success
		 */
		virtual OFBool setupCurlForCloudServiceType(const OFString& accessToken ) = 0;

		/**
		 * Get the id toke from access for the query access.
		 *
		 * @param access_token - accesss token
		 * @param audience - audience parameter for id token
		 * @param serviceAccountEmail - service account mail
		 *
		 * @return the id token
		 */
		OFString getIdTokenFromAccessToken( const OFString& access_token,
				const OFString& audience,
				const OFString& serviceAccountEmail );

		/**
		 * Get the property value from device configuration
		 *
		 * @param propertyName - property name
		 *
		 * @return the value
		 */
		OFString getDeviceProperty( const OFString& propertyName );

		/** Destructor */
		~GCPConnector();

	protected:

		/** GCP config for the device */
		IMProperties m_gcpConfig;
		/** Curl object for request*/
		CURL* m_curl;
		/** Curl headers*/
		struct curl_slist* m_curlHeaders;
		/** Device configuration path*/
		OFString m_deviceConfigPath;
		/** Query service base url*/
		OFString m_baseURL;
		/** Gcp Credentials for authentication*/
		std::shared_ptr<google::cloud::Credentials> m_credentials;
		OFBool initNetwork();
		/** Refresh the token for the query or gcp communications*/
		OFBool refreshToken();
		/** Destroy the network for the query or gcp communications*/
		void destroyNetwork();
		/** To encode the special character in mail id */
		OFString encodeMailIdSymbol( const OFString& email );

};

} // namespace erad::gcp

#endif /* IMUTILS_INCLUDE_IMUTILS_GCPCONNECTOR_H_ */
