/*
 * Component: utils
 *
 * Description:
 *    Header file for IMProperties class.
 *    IMProperties is a wrapper class for reading properties from a property file
 *
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef __IM_PROPERTY_H_INCLUDED__
#define __IM_PROPERTY_H_INCLUDED__

#include <list>
using namespace std;

//class IMProperty;
/**
 * Helper class for holding one property
 */
class IMProperty {
	private:
		char * m_name;
		char * m_value;

		char * copy( const char * str );

	public:

		IMProperty( const char * name, const char * value );
		IMProperty( const IMProperty& orig );

		~IMProperty();

		inline const char * getName() const {
			return m_name;
		}

		inline const char * getValue() const {
			return m_value;
		}

		void setValue( const char * value );
};

/**
 * Class for reading property files.
 * A property file is supposed to contain property - value pairs in 
 * propertyName="value"
 * format per lines.
 *
 * Lines starting with '#' are supposed to be comments and will be skipped.
 */
class IMProperties {
	protected:
		/** List of properties stored in the property file */
		list< IMProperty* > m_list;

		/**
		 * Finds the first non white spaces character in the argument string and returns its pointer.
		 * The argument string is supposed to be '\0' terminated
		 *
		 * @param str skip white spaces in this string
		 *
		 * @return pointer to the first non white space character in the string
		 */
		char * skipWhiteSpaces( char * str );

		/**
		 * Trims whitespaces from the argument string.
		 * The argument string will be modified: ending whitespaces will be replaced by '\0'.
		 * The return value will point the the first non whitespace character in the argument string.
		 *
		 * @param str string to be trimmed
		 *
		 * @return pointer to the first non white space character in the argument string
		 */
		char * trim( char * str );

		/**
		 * Replaces all occurences of src to dest in the string str.
		 * The str string will be modified!!
		 *
		 * @param str   string to replace chars in
		 * @param src   character to be replaced
		 * @param dest  replacing character
		 */
		void replace( char * str, char src, char dest );

		/**
		 * Returns the next value from str delimited by delim.
		 * The value is supposed to be start at the very first position and delimited by delim.
		 * The str string will be modified!!
		 *   The delimiter character will be replaced by a '\0' character.
		 *   The str pointer will be set to the first character after the delimiter found.
		 *
		 * @param str    [in|out]return value from this string
		 * @param delim  value delimiter
		 *
		 * @return pointer to the value in str or NULL if no delimiter found
		 */
		char * getNextValue( char *& str, char delim );

	public:

		/**
		 * Constructor for creating a new IMProperties object.
		 */
		IMProperties();

		/**
		 * Copy constructor.
		 *
		 * @param orig the original IMProperties object to copy
		 */
		IMProperties( const IMProperties& orig );

		/**
		 * Destructor for destroying the IMProperties object.
		 */
		~IMProperties();

		/**
		 * Loads properties from a property file.
		 *
		 * @param filename  name of the property file to be loaded
		 *
		 * @return true on success or false if an error occured
		 */
		bool load( const char * filename );

		/**
		 * Writes the properties into a property file
		 *
		 * @param filename  name of the property file to be written
		 *
		 * @return true on success or false if an error occured
		 */
		bool write( const char * filename );

		/**
		 * Returns the value of a property.
		 *
		 * @param name   name of the property to be returned
		 *
		 * @return property value or NULL if there is no such property
		 */
		const char * getProperty( const char * name ) const;

		/**
		 * Sets the value of a property.
		 * The property will be added to the property list if not yet exists.
		 * Neither name nor value can be NULL.
		 *
		 * @param name   name of the property to be set.
		 * @param value  new value for the property.
		 */
		void setProperty( const char * name, const char * value );

		/**
		 * Removes a property from the property list.
		 *
		 * @param name   name of the property to be removed
		 */
		void remove( const char * name );

		/**
		 * Removes all properties from the property list.
		 */
		void clear();
};

#endif // __IM_PROPERTY_H_INCLUDED__
