#ifndef IMUTILS_INCLUDE_IMUTILS_INTRACOMCLIENT_H_
#define IMUTILS_INCLUDE_IMUTILS_INTRACOMCLIENT_H_

#include <grpcpp/grpcpp.h>
#include "intracom.grpc.pb.h"
#include <google/protobuf/map.h>
#include <map>

#include "imutils/RoleConfig.h"

namespace erad::intracom {


/**
 * Encapsulates the communication layer for executing an intracom command.
 * The communication layer is an established gRPC channel either to a remote or the local device.
 */
class Courier {
	public:
		/* Disable the default constructor*/
		Courier() = delete;

		/**
		 * Creates a new courier to the server.
		 * The server expected to be in the @code{host}:@code{port} format,
		 * where @code{host} is the hostname or the IP address of the server, and
		 * @code{port} is the intracom port.
		 *
		 * @param role the role of the server in the farm
		 * @param server the server
		 */
		Courier( erad::farm::Role role, const std::string& server );

		/* Use the default copy and move constructor */

		Courier( const Courier& orig ) = default;
		Courier( Courier&& orig ) = default;

		/**
		 * Destroys the courier
		 */
		~Courier();

		/* Use the default copy and move constructor */

		Courier& operator=( const Courier& ) = default;
		Courier& operator=( Courier&& ) = default;

		/**
		 * Gets the gRpc channel to the server
		 *
		 * @return the gRpc channel to the server
		 */
		std::shared_ptr<grpc::Channel> getChannel() const {
			return channel;
		}

		/**
		 * Gets the gRpc channel state
		 *
		 * @return the gRpc channel state
		 */
		bool isValid() const;

		/**
		 * Gets the server
		 *
		 * @return the server in @code{host}:@code{port} format
		 */
		const std::string& getServer() const {
			return server;
		}

		/**
		 * Creates a stub for executing a command on the channel. Also sets meta headers.
		 *
		 * @param stub output parameter for the stub created for executing a command
		 * @param context output parameter for setting the meta data to be passed in the rpc call
		 */
		void fillStub( std::unique_ptr<rpc::IntraCom::Stub>& stub, ::grpc::ClientContext& context ) const;


	private:

		/// The role of the server in the farm;
		erad::farm::Role role;

		/// The server (host + port)
		std::string server;

		/// gRpc channel to the server
		std::shared_ptr<grpc::Channel> channel;
};


/**
 * Custom exception for IntraCom related errors
 */
class ICException : public std::exception {
	public:

		/**
		 * Creates a new ICException object.
		 *
		 * @param message error message to be returned by @code{what()}
		 */
		ICException( const std::string& message ) noexcept;

		/**
		 * Destroys the ICException object
		 */
		virtual ~ICException();

		/**
		 * @copydoc std::exception::what()
		 */
		virtual const char* what() const noexcept {
			return message.c_str();
		}

	private:
		/// The error message
		std::string message;
};


/**
 * Manage couriers and channels to the farm servers.
 * Provides a cache for the couriers and channels to the farm servers.
 */
class ChannelMgr {
	public:

		/**
		 * Creates a new ChannelMgr object using the provided farm configuration.
		 *
		 * @param roleConfig the farm configuration
		 */
		ChannelMgr( std::shared_ptr<erad::farm::RoleConfig> roleConfig );

		/**
		 * Destroys the ChannelMgr object.
		 */
		~ChannelMgr();

		/**
		 * Gets a courier to the server.
		 * Returns the existing courier to the server or creates a new one if there is no courier to the server yet.
		 *
		 * @param role the role of the server (may be @code{Role::Unknown})
		 * @param server the server in @code{host}:@code{port} format
		 * @return the courier to the server
		 */
		std::shared_ptr<Courier> getCourierFor( erad::farm::Role role, const std::string& server );

		/**
		 * Gets a courier for a particular role. The courier will be selected by the following rules.
		 * If the local server provides the requested role, then a courier to the localhost (127.0.0.1) will be returned.
		 * If there is a single providers for the role, then a courier to the particular server will be returned.
		 * If there are multiple providers for the role, then a courier to the load balancer will be returned.
		 *
		 * @param role the requested role
		 * @param retry if the channel is down, try to reconnect
		 * @return the courier to the role
		 * @throw RoleException if an error occurs while loading the role configuration or if getting information about the particular role is not supported.
		 */
		std::shared_ptr<Courier> getAnyCourierFor( erad::farm::Role role, bool retry = true );

		/**
		 * Gets couriers for all servers providing the particular role in the farm.
		 * When there are multiple providers for the role, then couriers connected directly to the servers will be returned (vs the load balancer)
		 *
		 * @param role the requested role
		 * @return couriers to all servers providing the role
		 */
		std::vector<std::shared_ptr<Courier>> getAllCouriersFor( erad::farm::Role role );

		/**
		 * Reconnect the server
		 *
		 * @param role the role of the server (may be @code{Role::Unknown})
		 * @param server reconnected server 
		 * @return the courier to the server
		 */
		std::shared_ptr<Courier> reconnectCourier( erad::farm::Role role, const std::string& server );

		/**
		 * Reload role config
		 */
		void reloadRoleConfig();

	private:

		/// Mutex for data writing
		std::recursive_mutex dataMutex;

		/// Farm configuration
		std::shared_ptr<erad::farm::RoleConfig> roleConfig;

		/// Cache for the couriers
		std::map<std::string, std::shared_ptr<Courier>> couriers;
};



/**
 * Client implementation to call intracom services and process responses
 */
class IntraComClient {
	public:
		/**
		 * Creates a new IntraComClient
		 */
		IntraComClient();

		/**
		 * Issues a hello request to a server in the farm.
		 *
		 * @param courier a courier to the server to call
		 * @param user user to be passed in the request
		 * @return reply received in the response
		 *
		 * @throws ICException if an error occurs
		 */
		static std::string hello( const Courier& courier, const std::string& user ) noexcept ( false );

		void getPbs( const Courier& courier, const std::string& viewSessionID, std::string& streamerID, std::string& oper,
				std::map<std::string, std::string>& cacheMap, std::string& replyPbs );

		void getViewerProfile( const Courier& courier, const std::string& sessionID, std::string& profile );
		bool putViewerProfile( const Courier& courier, const std::string& sessionID, std::string& profile );

		bool startStorescpReg( const Courier& courier,
				const std::string& study,
				const std::string& file,
				rpc::ICRegMode reindexMode,
				const std::string& tempFile,
				const std::string& mount,
				const std::string& hash,
				rpc::ICRunMode runMode,
				bool isDoubleCheck,
				bool isStartedFromObsolete,
				bool hasPixels,
				bool hasEditChanges,
				const std::string& studyDate,
				const std::string& senderDevice,
				const std::string& senderIP,
				const std::string& senderAE,
				const std::string& delta,
				const std::string& jobId,
				const std::string& subJobId,
				const std::string& childId,
				const std::string& jobInfo,
				int relativePriority
				) noexcept ( false );


		std::string addTask(
				const Courier& courier,
				const std::string& priorityType,
				const std::string& command,
				rpc::ICTaskType type,
				rpc::ICTimeMode timeMode,
				int timeSeconds,
				rpc::ICCollapseType collapseType,
				const std::string& deltaType,
				const std::string& jobId,
				const std::string& subJobId,
				const std::string& jobInfo,
				int relPriority,
				rpc::ICRunMode runMode );

		/**
		 * Issues a mapObjectTable request to re-map object table entries for a study.
		 *
		 * @param courier a courier to the server to call
		 * @param study the study instance UID
		 * @return @code{true} if and only if the object table entries get successfully re-mapped for the study
		 *
		 * @throws ICException if an error occurs
		 */
		bool mapObjectTable( const Courier& courier,
				const std::string& study );

		/**
		 * Issues a mapObjectTable request to remove (un-map) object table entries for a study.
		 *
		 * @param courier a courier to the server to call
		 * @param study the study instance UID
		 * @return @code{true} if and only if the object table entries get successfully removed (un-mapped) for the study
		 *
		 * @throws ICException if an error occurs
		 */
		bool unmapObjectTable( const Courier& courier,
				const std::string& study, int timeout );

};

} // namespace erad::intracom

#endif /* IMUTILS_INCLUDE_IMUTILS_INTRACOMCLIENT_H_ */
