/*
 * RefcountDB.h
 *
 *  Created on: Mar 28, 2022
 *      Author: medsrv
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_LOCKDB_H_
#define IMUTILS_INCLUDE_IMUTILS_LOCKDB_H_

#include <string>
#include <list>
#include "imutils/DBConnector.h"

namespace erad::db {


/**
 * Interface to manage reference counters
 */
class LockDB {

	public:
		virtual ~LockDB() {
		}

		/**
		 * Trys to add a new lock entry ("locking attempt")
		 *
		 * @param id the lock id
		 * @param resource the resource (name) to be locked
		 * @param actor entity requesting the lock
		 * @param type lock type
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return true if entry has been created ("locking attempt succeeded"), false otherwise
		 */
		virtual bool lockFirstTime( const std::string& id, const std::string& resource, const std::string& actor, const int type ) = 0;

		/**
		 * Removes lock with specific id ("unlock")
		 *
		 * @param id the lock id

		 * @throw DBException if an DB error occurs
		 *
		 * @return true if entry has been found and removed, false otherwise
		 */
		virtual bool releaseLock( const std::string& id ) = 0;
};


/**
 * Reference counter implementation managing reference counters in MySQL database tables.
 */
class LockDBSql : public LockDB, public PSCache {

	public:

		/**
		 * Constructs a new LockDBSql object to prepare and execute MySQL queries
		 *
		 * @param pDB connector to the MySQL database
		 */
		LockDBSql( DBConnectorSql& pDB );
		virtual ~LockDBSql();

		/**
		 * @copydoc LockDB::lockFirstTime(const std::string&,const std::string&,std::string&,const int)
		 */
		virtual bool lockFirstTime( const std::string& id, const std::string& resource, const std::string& actor, const int type ) noexcept ( false )
				override;

		/**
		 * @copydoc LockDB::releaseLock(const std::string&)
		 */
		virtual bool releaseLock( const std::string& id ) noexcept ( false )
				override;

	private:

		std::string lockTbl;
		std::string objectTbl;

		PsQuery lockFirstTimeQuery; // NEW_LOCK( () -> "insert into " + TableEnum.EPLOCKTBL.getTableSQLName() + " (LCKID,LCKRESOURCE,LCKACTOR,LCKTYPE,LCKTIME,LCKCOUNTER) values (?,?,?,?,NOW(),1)" )
		PsQuery releaseLockQuery; // REM_LOCK( () -> "delete from " + TableEnum.EPLOCKTBL.getTableSQLName() + " where  LCKID=? and LCKCOUNTER<2" )
};

} // namespace

#endif /* IMUTILS_INCLUDE_IMUTILS_LOCKDB_H_ */
