/*
 * Interface definition of accessing and managing information stored in
 * different database tables
 *
 * Also implements the interface for MySQL.
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_MISCDB_H_
#define IMUTILS_INCLUDE_IMUTILS_MISCDB_H_

#include <string>
#include <list>
#include <vector>
#include "imutils/DBConnector.h"

namespace erad::db {

/* ********************************************
 *
 * Helper bundles to pass/return information by the interface methods
 *
 * ********************************************/


/**
 * A bundle to hold DICOM related information of a field, especially a custom field.
 */
class DicomInfo {
	public:

		/// The DICOM tag of the field
		std::string dicomTag;
		/// The property name of a user defined custom field
		std::string customFieldName;

		/**
		 * Default constructor creating an empty DicomInfo object
		 */
		DicomInfo() noexcept {
		}

		/**
		 * Constructor to create a new DicomInfo object
		 * @param dicomTag the DICOM tag of the field
		 * @param customFieldName the property name of a user defined custom field
		 */
		DicomInfo( const std::string& dicomTag, const std::string& customFieldName ) noexcept
				: 	dicomTag( dicomTag ),
					customFieldName( customFieldName ) {
		}


		/* Copy and move constructors are disabled */

		DicomInfo( const DicomInfo& orig ) = delete;
		DicomInfo( DicomInfo&& orig ) = delete;
		
		~DicomInfo() {
		}

		/* Copy and move assignment operators are also disabled */

		DicomInfo& operator=( const DicomInfo& ) = delete;
		DicomInfo& operator=( DicomInfo&& ) = delete;
};


/**
 * A bundle to hold information for a device
 */
class DeviceInfo {
	public:

		/// The device ID
		std::string id;
		/// The device type
		std::string type;

		/**
		 * Constructor to create a new DeviceInfo object
		 * @param id the device ID
		 * @param type the device type
		 */
		DeviceInfo( const std::string& id, const std::string& type )
				: 	id( id ),
					type( type ) {
		}

		/* Copy and move constructors are disabled */

		DeviceInfo( const DeviceInfo& orig ) = delete;
		DeviceInfo( DeviceInfo&& orig ) = delete;

		~DeviceInfo() {
		}

		/* Copy and move assignment operators are also disabled */

		DeviceInfo& operator=( const DeviceInfo& ) = delete;
		DeviceInfo& operator=( DeviceInfo&& ) = delete;
};


/**
 * A bundle to hold (DICOM) connection information to a device
 */
class ConnectionInfo {
	public:

		/// The IP address of the device
		std::string ipAddress;
		/// DICOM port
		int port;
		/// Protocol
		std::string protocol;
		/// AE title
		std::string aeTitle;

		/**
		 * Default constructor creating an empty ConnectionInfo object
		 */
		ConnectionInfo()
				: port( 104 ) {
		}

		/**
		 * Constructor to create a new ConnectionInfo object
		 * @param ipAddress the IP address
		 * @param port the DICOM port
		 * @param protocol the protocl ('S' for TLS, 'T' for TCP)
		 * @param aeTitle the AE title
		 */
		ConnectionInfo( const std::string& ipAddress, int port, const std::string& protocol,
				const std::string& aeTitle )
				: 	ipAddress( ipAddress ),
					port( port ),
					protocol( protocol ),
					aeTitle( aeTitle ) {
		}

		/* Copy and move constructors are disabled */

		ConnectionInfo( const ConnectionInfo& orig ) = delete;
		ConnectionInfo( ConnectionInfo&& orig ) = delete;

		virtual ~ConnectionInfo() {
		}

		/* Copy and move assignment operators are also disabled */

		ConnectionInfo& operator=( const ConnectionInfo& ) = delete;
		ConnectionInfo& operator=( ConnectionInfo&& ) = delete;
};


/**
 * A bundle to hold information for a log field value
 */
class LogField {
	private:

		/// database column name of the log field
		std::string dbname;
		/// the value to be logged
		std::string value;

	public:

		/**
		 * Constructor to create a new LogField object
		 * @param dbname database column name of the log field
		 * @param type the value to be logged
		 */
		LogField( const std::string& dbname, const std::string& value )
				: 	dbname( dbname ),
					value( value ) {
		}

		/* Default, copy and move constructors are disabled */

		LogField() = delete;
		LogField( const LogField& orig ) noexcept = default;
		LogField( LogField&& orig ) noexcept = default;

		~LogField() {
		}

		/* Copy and move assignment operators are also disabled */

		LogField& operator=( const LogField& ) noexcept = default;
		LogField& operator=( LogField&& ) noexcept = default;

		inline const std::string& getDbName() const {
			return dbname;
		}

		inline const std::string& getValue() const {
			return value;
		}
};


/* ********************************************
 *
 * Interface
 *
 * ********************************************/


/**
 * Interface definition to access and manage information stored in
 * miscellaneous database tables
 */
class MiscDB {

	public:
		virtual ~MiscDB() {
		}

		/* Table descriptors */

		/**
		 * Gets the DICOM tag(s) of a field.
		 * Most fields have a single DICOM tag, but date/time fields - or image size - consist of multiple tags.
		 *
		 * Used by edit to read merge.cfg for hard/soft edit configuration.
		 *
		 * @param colid [in] the field ID
		 * @param dicomTags [out] output parameter for the DICOM tags
		 *
		 * @throw DBException if a DB error occurs
		 */
		virtual void getDicomTag( const std::string& colid, std::vector<std::string>& dicomTags /* out */) noexcept ( false ) = 0;

		/**
		 * Gets DICOM information (tag, custom name) for a field by he database name.
		 *
		 * Used by Order data filler to look up DICOM tags for database fields when pre-filling DICOM datasets
		 * with study level and demographic data taken from the database.
		 *
		 * @param dbname [in] the database column name of a field
		 * @param dicomInfo [out] output parameter for the dicom information of the field
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the field was found in the descriptor tables
		 */
		virtual bool getDicomInfoByDbname( const std::string& dbname, DicomInfo& dicomInfo /* out */) noexcept ( false ) = 0;

		/**
		 * Gets the database field name for a DICOM tag.
		 * The return value indicates whether or not a database field exists for the DICOM tag.
		 *
		 * Used by pb-scp when serving DICOM Q/R and MWL C-Find requests to look up database field names
		 * for DICOM tags recieved in the query request.
		 *
		 * @param dicomTag [in] the DICOM tag to lookup
		 * @param table [in] the database table to check
		 * @param dbname [out] output parameter for the database column name
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the field is mapped to a database field
		 */
		virtual bool getDbnameByDicomTag( const std::string& dicomTag, const std::string& table, std::string& dbname /*out*/)
				noexcept ( false ) = 0;

		/**
		 * Gets the database field name for a DICOM date (DA) or time (TM) tag.
		 * The return value indicates whether or not a database field exists for the DICOM tag.
		 *
		 * Used by pb-scp when serving DICOM Q/R and MWL C-Find requests to look up database field names
		 * for DICOM tags recieved in the query request.
		 *
		 *
		 * @param dicomTag [in] the DICOM tag to lookup
		 * @param table [in] the database table to check
		 * @param isDate [in] true indicates the tag is a date (DA) field (otherwise it is expected to be a time (TM) field)
		 * @param dbname [out] output parameter for the database column name
		 * @param dicomTags [out] output parameter for the dicom tags for both (date and time) fields
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the field is mapped to a database field
		 */
		virtual bool getDbnameByDateTimeTag( const std::string& dicomTag, const std::string& table, bool isDate,
				std::string& dbname /*out*/, std::string& dicomTags /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets the DICOM tag and the database type for a field by the database name.
		 *
		 * Used for processing (order-, patient-) matching rules. The rules use database column names
		 *
		 * @param dbname [in] the database column name of a field
		 * @param dicomTag [out] output parameter for the DICOM tag of the field
		 * @param type [out] output parameter for the database field type of the field
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the field was found in the descriptor tables
		 */
		virtual bool getTagAndTypeByDbname( const std::string& dbname, std::string& dicomTag /* out */,
				std::string& type /* out */) noexcept ( false ) = 0;

		/* Transactions */

		/**
		 * Gets the status of a transaction.
		 *
		 * Used by the transaction manager.
		 *
		 * @param transaction [in] the transaction ID
		 * @param status [out] output parameter for the transaction status
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the transaction was found
		 */
		virtual bool getTransactionState( const std::string& transaction, std::string& status /* out */) noexcept ( false ) = 0;

		/**
		 * Gets the status and the result of a transaction.
		 * The result is only set when the status indicates the transaction is completed.
		 *
		 * Used by the transaction manager.
		 *
		 * @param transaction [in] the transaction ID
		 * @param status [out] output parameter for the transaction status
		 * @param answer [out] output parameter for the result of the transaction
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the transaction was found
		 */
		virtual bool getTransactionAnswer( const std::string& transaction, std::string& status /* out */,
				std::string& answer /* out */) noexcept ( false ) = 0;

		/**
		 * Creates or updates a transaction.
		 *
		 * Used by the transaction manager.
		 *
		 * @param transaction [in] the transaction ID
		 * @param timestamp [in] new timestamp to be set for the transaction
		 * @param status [in] new status for the transaction
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return the number of transaction records changed
		 */
		virtual size_t updateTransaction( const std::string& transaction, const std::string& timestamp, const std::string& status )
				noexcept ( false ) = 0;

		/**
		 * Completes a transaction.
		 * Also sets the result for the transaction
		 *
		 * @param transaction [in] the transaction ID
		 * @param timestamp [in] new timestamp to be set for the transaction
		 * @param status [in] new status for the transaction
		 * @param answer [in] the result for the transaction
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return the number of transaction records changed
		 */
		virtual size_t completeTransaction( const std::string& transaction, const std::string& timestamp, const std::string& status,
				const std::string& answer ) noexcept ( false ) = 0;

		/* Devices */

		/**
		 * Gets the list of the auto-forward devices.
		 *
		 * Used by StoreState to  fill the ForwardDevices attribute when creating a storestate.rec file for a study.
		 *
		 * @param devices [out] output parameter to return the Device ID and the Device Type of the Auto-forward devices
		 *
		 * @throw DBException if an DB error occurs
		 */
		virtual void getAfwdDevices( std::list<DeviceInfo>& devices /* out */) noexcept ( false ) = 0;

		/**
		 * Gets connection information for a device.
		 * The return value indicates whether the device exists or not.
		 *
		 * Used by MCS retrieve to get connection information to the PACS server
		 *
		 * @param deviceId [in] the device ID
		 * @param info [out] output parameter for the connection information
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return true if and only if the device was found in the database
		 */
		virtual bool getConnectionInfo( const std::string& deviceId, ConnectionInfo& info /* out */) noexcept ( false ) = 0;


		/* Logging to database */

		/**
		 * Collects registered callbacks for a particular log action
		 *
		 * @param action [in] the log action
		 * @param callbacks [out] output parameter for the registered callbacks for the action
		 *
		 * @throw DBException if an DB error occurs
		 */
		virtual void getCallbacks( const std::string& action, std::list<std::string>& callbacks /*out*/) noexcept ( false ) = 0;

		/**
		 * Creates a log table entry using the provided field values
		 *
		 * @param fields [in] the field values for the log entry
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return the number of log entries created (1 or 0)
		 */
		virtual size_t operLog( const std::vector<LogField>& fields ) noexcept ( false ) = 0;

		/* MCS */

		/**
		 * Inserts a new step/substep to the MCS progress table
		 *
		 * Used by MCS retrieve to add progress steps for the object(s) to retreive to the MCS progress table.
		 *
		 * @param mediaId [in] the media session identifier
		 * @param stepId [in] identifies the step to insert
		 * @param subStepId [in] identifies the sub-step to insert
		 * @throw DBException if an DB error occurs
		 *
		 * @return the number of progress step records created (1 or 0)
		 */
		virtual size_t insertMcsProgress( const std::string& mediaId, const std::string& stepId, const std::string& subStepId )
				noexcept ( false ) = 0;

		/**
		 * Gets the inbound relative priority for the device
		 *
		 * @param	deviceId the ID of the device
		 * @throw DBException if an DB error occurs
		 *
		 * @return	the inbound relative priority for the device
		 */
		virtual int getDeviceInboundRelpriority( const std::string& deviceId ) noexcept ( false ) = 0;

};


/* ********************************************
 *
 * Implementation
 *
 * ********************************************/


/**
 * Implementation of the MiscDB interface for MySQL
 */
class MiscDBSql : public MiscDB, public PSCache {

	public:

		/**
		 * Constructs a new MiscDBSql object to prepare and execute MySQL queries
		 *
		 * @param pDB connector to the MySQL database
		 */
		MiscDBSql( DBConnectorSql& pDB );
		virtual ~MiscDBSql();

		/* Table descriptors */

		/**
		 * @copydoc MiscDB::getDicomTag(const std::string&,std::vector<std::string>&)
		 */
		virtual void getDicomTag( const std::string& colid, std::vector<std::string>& dicomTags /* out */) noexcept ( false )
				override;

		/**
		 * @copydoc MiscDB::getDicomInfoByDbname(const std::string&,DicomInfo&)
		 */
		virtual bool getDicomInfoByDbname( const std::string& dbname, DicomInfo& dicomInfo /* out */) noexcept ( false ) override;

		/**
		 * @copydoc MiscDB::getDbnameByDicomTag(const std::string&,const std::string&,std::string&)
		 */
		virtual bool getDbnameByDicomTag( const std::string& dicomTag, const std::string& table, std::string& dbname /*out*/)
				noexcept ( false ) override;

		/**
		 * @copydoc MiscDB::getDbnameByDateTimeTag(const std::string&,const std::string&,bool,std::string&,std::string&)
		 */
		virtual bool getDbnameByDateTimeTag( const std::string& dicomTag, const std::string& table, bool isDate,
				std::string& dbname /*out*/, std::string& dicomTags /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc MiscDB::getTagAndTypeByDbname(const std::string&,std::string&,std::string&)
		 */
		virtual bool getTagAndTypeByDbname( const std::string& dbname, std::string& dicomTag /* out */,
				std::string& type /* out */) noexcept ( false ) override;

		/* Transactions */

		/**
		 * @copydoc MiscDB::getTransactionState(const std::string&,std::string&)
		 */
		virtual bool getTransactionState( const std::string& transaction, std::string& status /* out */) noexcept ( false ) override;

		/**
		 * @copydoc MiscDB::getTransactionAnswer(const std::string&,std::string&,std::string&)
		 */
		virtual bool getTransactionAnswer( const std::string& transaction, std::string& status /* out */,
				std::string& answer /* out */) noexcept ( false ) override;

		/**
		 * @copydoc MiscDB::updateTransaction(const std::string&,const std::string&,const std::string&)
		 */
		virtual size_t updateTransaction( const std::string& transaction, const std::string& timestamp, const std::string& status )
				noexcept ( false ) override;

		/**
		 * @copydoc MiscDB::completeTransaction(const std::string&,const std::string&,const std::string&,const std::string&)
		 */
		virtual size_t completeTransaction( const std::string& transaction, const std::string& timestamp, const std::string& status,
				const std::string& answer ) noexcept ( false ) override;

		/* Devices */

		/**
		 * @copydoc MiscDB::getAfwdDevices(std::list<std::tuple<std::string,std::string>>&)
		 */
		virtual void getAfwdDevices( std::list<DeviceInfo>& devices /* out */) noexcept ( false ) override;

		/**
		 * @copydoc MiscDB::getConnectionInfo(const std::string&,ConnectionInfo&)
		 */
		virtual bool getConnectionInfo( const std::string& deviceId, ConnectionInfo& info /* out */) noexcept ( false )
				override;

		/* Logging to database */

		/**
		 * @copydoc MiscDB::getCallbacks(const std::string&,std::list<std::string>&)
		 */
		virtual void getCallbacks( const std::string& action, std::list<std::string>& callbacks /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc MiscDB::operLog(const std::vector<LogField>&)
		 */
		virtual size_t operLog( const std::vector<LogField>& fields ) noexcept ( false ) override;

		/* MCS */

		/**
		 * @copydoc MiscDB::insertProgress(const std::string&,const std::string&,const std::string&)
		 */
		virtual size_t insertMcsProgress( const std::string& mediaId, const std::string& stepId, const std::string& subStepId )
				noexcept ( false ) override;

		/**
		 * @copydoc MiscDB::getDeviceInboundRelpriority(const std::string&)
		 */
		virtual int getDeviceInboundRelpriority( const std::string& deviceId ) noexcept ( false ) override;

	private:

		std::string fieldsTbl;
		std::string typesTbl;
		std::string tablesTbl;
		std::string transactionTbl;
		std::string targetTbl;
		std::string devservTbl;
		std::string callbackRegistryTbl;
		std::string callbackUrlTbl;
		std::string logTbl;
		std::string mcsProgressTbl;

		PsQuery dicomTagByColid;
		PsQuery dicomInfoByDbname;
		PsQuery dbnameByDicomTag;
		PsQuery dbnameByDateTimeTag;
		PsQuery tagAndTypeByDbname;
		PsQuery transactionState;
		PsQuery transactionAnswer;
		PsQuery transactionInProcess;
		PsQuery transactionDone;
		PsQuery afwdDevices;
		PsQuery connectionInfo;
		PsQuery collectCallbacks;
		PsQuery mcsProgress;
		PsQuery deviceInboundRelpri;
};

} // namespace

#endif /* IMUTILS_INCLUDE_IMUTILS_MISCDB_H_ */
