/**
 * OOP wrapper classes around MySQL C API
 */
#ifndef IMUTILS_INCLUDE_IMUTILS_EPSQL_H_
#define IMUTILS_INCLUDE_IMUTILS_EPSQL_H_

#include <string>
#include <sstream>
#include <mysql.h>
#include "mysqld_error.h"
#include <vector>
#include <mutex>

namespace erad {
namespace db {

class SqlConnection;
class SqlStatement;


/**
 * Controls whether or not to set the is_unsigned flag when binding SQL parameters or results of an integral type.
 */
enum class SignedFlag {
	/** signedness of the parameter/result is undefined (not an integral type) */
	UNDEFINED,
	/** parameter/result is a signed integral type */
	SIGNED,
	/** parameter/result is an unsigned integral type */
	UNSIGNED
};


/**
 * Exceptions used to indicate MySQL errors
 */
class SqlException : public std::exception {
		/// the MySQL error code
		unsigned int errorCode;

		/// the application message
		std::string appMessage;

		/// the error message
		std::string mysqlMessage;

		/// the message combined from the application and mysql messages
		std::string message;

	public:

		/**
		 * Constructs a new SQL Exception having the error code and the error message provided by the mysql library.
		 *
		 * @param errorCode the error code
		 * @param message the error message
		 */
		SqlException( unsigned int errorCode, const std::string& appMessage, const std::string& mysqlMessage ) noexcept;

		/* Use the default copy and move constructors */

		SqlException( const SqlException& orig ) noexcept = default;
		SqlException( SqlException&& orig ) noexcept = default;

		virtual ~SqlException();

		/* Use the default copy and move assignments */

		SqlException& operator=( const SqlException& ) noexcept = default;
		SqlException& operator=( SqlException&& ) noexcept = default;

		/**
		 * @copydoc std::exception::what()
		 */
		virtual const char* what() const noexcept override {
			return message.c_str();
		}

		/**
		 * Gets the application (contextual) message.
		 * @return the application message
		 */
		inline const std::string& getAppMessage() const noexcept {
			return appMessage;
		}

		/**
		 * Gets the MySQL error message
		 * @return the MySQL error message
		 */
		inline const std::string& getMysqlMessage() const noexcept {
			return mysqlMessage;
		}

		/**
		 * Gets the MySQL error code
		 * @return the MySQL error code
		 */
		inline unsigned int getErrorCode() const noexcept {
			return errorCode;
		}

		/**
		 * Checks if the MySQL error indicates a condition to reset the connector
		 * (close the connections and drop existing prepared statements).
		 *
		 * Checks if the connection has timed out or get dropped.
		 *
		 * @return true if and only if the connector need to be reset.
		 */
		inline bool needReconnect() const noexcept {
			return ( errorCode == ER_CLIENT_INTERACTION_TIMEOUT
					|| errorCode == CR_SERVER_GONE_ERROR || errorCode == CR_SERVER_LOST
			);
		}

		/**
		 * Checks if the MySQL error indicates a deadlock issue
		 * these are safe to just retry.
		 *
		 * HPS-211: handle mysql deadlock exceptions
		 *
		 * @return true if and only if the error is of type deadlock
		 */
		inline bool canRetryAfterDeadlock() const noexcept {
			return ( errorCode == ER_LOCK_DEADLOCK	|| errorCode == ER_XA_RBDEADLOCK
					|| errorCode == ER_USER_LOCK_DEADLOCK || errorCode == ER_LOCKING_SERVICE_DEADLOCK
			);
		}

};

/**
 * A wrapper class for the SQL client to initialize/finalize the MySQL client library
 * and create SQL connections.
 *
 * There should be a single instance for each thread in the application.
 */
class Sql {

	public:

		/**
		 * Constructs a new SQL client for the main process or a thread.
		 * Initializes the MySQL client library.
		 *
		 * @param fromThread true indicates to initialize the MySQL for a thread
		 */
		Sql( bool fromThread = false );

		/**
		 * Destroys the SQL client.
		 * Finalizes the MySQL client library.
		 */
		~Sql();


		/**
		 * Creates a new SQL connection
		 *
		 * @throw SqlException if an error occurs
		 * @return the new connection
		 */
		SqlConnection* getConnectionEx() noexcept ( false );

		/**
		 * Creates a new SQL client for a new thread.
		 *
		 * @return the new SQL client or NULL if an error occurs
		 */
		inline Sql* sqlThread() {
			return new Sql( *this );
		}

		/**
		 * Checks if the SQL client could be successfully initialized.
		 *
		 * @return true if and only if the new SQL client was initialized successfully
		 */
		inline bool isGood() const {
			return good;
		}

		/**
		 * Returns the error message for the last operation (initialization or connection).
		 *
		 * @return the error message for the last operation
		 */
		inline const std::string& getError() const {
			return errorMsg;
		}

		/**
		 * Gets the reentrant mutex used to prevent running simultaneous queries on the same connection on multiple threads.
		 * The mutex is also used for creating connections and reconnecting when the connection is gone.
		 *
		 * @return the reentrant mutex
		 */
		inline std::recursive_mutex& getConMutex() {
			return conMutex;
		}


	protected:

		/**
		 * Constructor used by sqlThread() to create a new SQL client for a new thread
		 *
		 * @param orig the original SQL client (for the main thread)
		 */
		Sql( const Sql& orig );

		/*
		 * Disable the
		 * - move constructor
		 * - copy assignment
		 * - move assignment
		 */
		Sql( Sql&& ) = delete;
		Sql& operator=( const Sql& ) = delete;
		Sql& operator=( Sql&& ) = delete;

	private:

		/**
		 * Checks if an environment variable is set and stores to a string
		 *
		 * @param envName   the environment variable to set
		 * @param var       output parameter to store the value of the environment variable
		 * @param mandatory true indicates the environment variable is mandatory
		 *
		 * @return true if the the environment variable was set or is not mandatory.
		 */
		bool checkEnv( const char* envName, std::string& var, bool mandatory );

		/**
		 * Connects to a local SQL server through a unix socket.
		 *
		 * @throw SqlException if an error occurs
		 *
		 * @return the new SQL connection
		 */
		MYSQL* connectUnixPortEx() noexcept ( false );

		/**
		 * Connects to a remote SQL server through a TCP connection
		 *
		 * @throw SqlException if an error occurs
		 *
		 * @return the new SQL connection
		 */
		MYSQL* connectTCPEx() noexcept ( false );

		/** database name */
		std::string dbname;
		/** database user */
		std::string dbuser;
		/** unix port of a local server */
		std::string unixPort;
		/** TCP host of a remote server */
		std::string host;
		/** TCP port of a remote server */
		int port;
		/** true indicates this client is for a thread */
		bool isThread;
		/** true indicates this client has no error */
		bool good = true;
		/** error message for the last failed operation */
		std::string errorMsg;

		/// A reentrant mutex to prevent running simultaneous connects and reconnects in concurrent threads
		std::recursive_mutex conMutex;
};


/**
 * Bundle for a return field in a MySQL basic query.
 */
class SqlField {

		/// the field name
		std::string name;
		/// the field value
		std::string value;
		/// indicates whether the field value is NULL
		bool null;

	public:

		/**
		 * Creates a new SqlField with a NULL value
		 * @param name the field name
		 */
		SqlField( const std::string& name ) noexcept;

		/**
		 * Creates a new SqlField with a not NULL value
		 * @param name the field name
		 * @param value the value of the field
		 */
		SqlField( const std::string& name, const std::string& value ) noexcept;

		/* Default, copy and move constructors are disabled */

		SqlField() = delete;
		SqlField( const SqlField& orig ) = delete;
		SqlField( SqlField&& orig ) = default;

		~SqlField();

		/* Copy and move assignment operators are also disabled */

		SqlField& operator=( const SqlField& ) = delete;
		SqlField& operator=( SqlField&& ) = delete;

		/**
		 * Gets the name of the field.
		 * @return the name of the field
		 */
		const std::string& getName() const {
			return name;
		}

		/**
		 * Gets the value of the field.
		 * @note It is recommended to check if the value is NULL or not calling isNull() before getValue().
		 * @return the field value
		 */
		const std::string& getValue() const {
			return value;
		}

		/**
		 * Checks if the field has a NULL value.
		 * @return true if the field value is NULL
		 */
		bool isNull() const {
			return null;
		}
};


/**
 * The result set of a MySQL basic query returning rows.
 */
class SqlResult {
	public:
		/**
		 * Creates a new SqlResult object for the provided MySQL result set
		 * @param res the result set
		 */
		SqlResult( MYSQL_RES* res ) noexcept;

		/* Default, copy and move constructors are disabled */

		SqlResult() = delete;
		SqlResult( const SqlResult& orig ) = delete;
		SqlResult( SqlResult&& orig ) = delete;

		~SqlResult();

		/* Copy and move assignment operators are also disabled */

		SqlResult& operator=( const SqlResult& ) = delete;
		SqlResult& operator=( SqlResult&& ) = delete;

		/**
		 * Fetches the next row from the result set.
		 *
		 * @return true if the next row was successfully fetched
		 */
		bool fetch() noexcept;

		/**
		 * Gets the field values of the current row.
		 *
		 * @param record output parameter for the field values
		 *
		 * @note: should only be called after a successful fetch()
		 */
		void getRecord( std::vector<SqlField>& record /* out */) noexcept;

	private:

		/// the result set of the query
		MYSQL_RES* res;
		/// the current row in the result set
		MYSQL_ROW rec;
		/// meta information about the field in the result set
		MYSQL_FIELD* fields;
		/// number of fields in the result set
		unsigned int numFields;
};


/**
 * Wrapper class around a MySQL connection
 */
class SqlConnection {

		friend class Sql;

	public:

		/**
		 * Destroys the connection.
		 * Releases the resources used by the connection.
		 */
		~SqlConnection();

		/**
		 * Prepares an SQL statement to execute a query.
		 *
		 * @param query the query to execute
		 *
		 * @throw SqlException if an error occurs
		 *
		 * @return the prepared statement
		 */
		SqlStatement* prepareStatementEx( const std::string& query ) noexcept ( false );

		/**
		 * Returns the error message for the last query.
		 */
		inline const std::string& getError() const {
			return errorMsg;
		}

		/**
		 * Checks whether the last MySQL basic query executed by executeQuery() was interrupted or not.
		 *
		 * @return true if and only if the last basic query was interrupted.
		 */
		inline bool isInterrupted() const {
			return interrupted;
		}

		/**
		 * Executes a MySQL basic query returning rows. (select).
		 * When a timeout is provided, then executing the query fails if does not complete within the specified timeout-
		 *
		 * @param query the query to execute
		 * @param timeout timeout in seconds (default: 0 meaning infinite)
		 *
		 * @throw SqlException if an error occurs
		 *
		 * @return the result of the query or NULL if an error occurred
		 */
		SqlResult* executeQueryEx( const std::string& query, size_t timeout = 0 ) noexcept ( false );

		/**
		 * Escapes a string to be used in a MySQL basic query.
		 *
		 * @param str the string to be escaped
		 * @param encodedStr output parameter for the sql encoded string
		 */
		void escape( const std::string& str, std::string& encodedStr );

		/**
		 * Checks if the connection is still alive
		 *
		 * @return true if and only if the connection is still alive
		 */
		bool ping();

	protected:

		/**
		 * Constructs a new SQL connection.
		 * Called by Sql::getConnection()
		 *
		 * @param connection the MySQL connection
		 */
		SqlConnection( MYSQL* connection, std::recursive_mutex & conMutex );

		/*
		 * Disable the
		 * - default constructor
		 * - copy constructor
		 * - move constructor
		 * - copy assignment
		 * - move assignment
		 */
		SqlConnection() = delete;
		SqlConnection( const SqlConnection& ) = delete;
		SqlConnection( SqlConnection&& ) = delete;
		SqlConnection& operator=( const SqlConnection& ) = delete;
		SqlConnection& operator=( SqlConnection&& ) = delete;

	private:

		/** the MySQL connection */
		MYSQL* connection;
		/** error message for the last failed operation */
		std::string errorMsg;

		/// indicates whether the most recent query was interrupted or not
		bool interrupted;

		/// Reference to the reentrant mutex of the connection, to prevent running simultaneous queries in concurrent threads
		std::recursive_mutex &conMutex;
};


/**
 * Binding to a field to be returned by a particular SQL query.
 * This is a wrapper for MYSQL_BIND.
 */
class QueryResult {

		friend class QueryBindings;

	public:

		/**
		 * Constructs a new Query Result object to bind a return field of a particular query
		 *
		 * @param type       the type of the field
		 * @param buffer     a pointer to the buffer to store the value for the field
		 * @param bufferSize the size of the buffer
		 * @param signedFlag signed flag for the field (when the field is an integral type)
		 */
		QueryResult( enum_field_types type, void* buffer, size_t bufferSize, SignedFlag signedFlag = SignedFlag::UNDEFINED );

		/**
		 * Destroys the Query result
		 */
		~QueryResult();

		/*
		 * Disable the
		 * - default constructor
		 * - copy constructor
		 * - move constructor
		 * - copy assignment
		 * - move assignment
		 */
		QueryResult() = delete;
		QueryResult( const QueryResult& orig ) = delete;
		QueryResult( QueryResult&& ) = delete;
		QueryResult& operator=( const QueryResult& ) = delete;
		QueryResult& operator=( QueryResult&& ) = delete;

		/**
		 * @return true (1) if the field value is NULL
		 */
		inline /*my_*/bool isNull() const {
			return isnull;
		}

		/**
		 * @return true (1) if an error occurred when mapping value for the last SQL record
		 */
		inline /*my_*/bool isError() const {
			return error;
		}

		/**
		 * @return the actual length of the value in the current record
		 */
		inline size_t getValueLength() const {
			return valueLength;
		}

		/**
		 * Checks if the current value of a field returned by the SQL query was mapped successfully and is not NULL.
		 *
		 * @return true if and only if the current value was successfully mapped and is not NULL
		 */
		inline bool isGood() const {
			return !error && !isnull;
		}

	protected:

		/** the field type */
		enum enum_field_types type;
		/** pointer to the buffer to store the value */
		char* buffer;
		/** the size of the buffer */
		size_t bufferSize;
		/** the actual length of the current value */
		size_t valueLength;
		/** true indicates the current value is NULL */
		/*my_*/bool isnull;
		/* signed flag for the field (when the field is an integral type) */
		SignedFlag signedFlag;
		/** true indicates mapping the current value failed */
		/*my_*/bool error;
};


/**
 * Binding of query parameters and return fields for a prepared statement.
 */
class QueryBindings {
	public:

		/**
		 * Creates a new QueryBindings object.
		 *
		 * @param numParams the number of parameters to bind for the query
		 * @param numResults the number of results (return fields) to bind for the query
		 */
		QueryBindings( int numParams, int numResults = 0 );
		~QueryBindings();

		/*
		 * Disable the
		 * - default constructor
		 * - copy constructor
		 * - move constructor
		 * - copy assignment
		 * - move assignment
		 */
		QueryBindings() = delete;
		QueryBindings( const QueryBindings& orig ) = delete;
		QueryBindings( QueryBindings&& ) = delete;
		QueryBindings& operator=( const QueryBindings& ) = delete;
		QueryBindings& operator=( QueryBindings&& ) = delete;

		/**
		 * Binds a parameter to fill input values in the query.
		 *
		 * @param index index of the input value
		 * @param type the field type
		 * @param buffer a pointer to the buffer holding the parameter value
		 * @param bufferSize the buffer size
		 * @param valueLength a pointer to the length of the actual value
		 * @param isNull a pointer to a boolean variable indicating whether or not the value is NULL.
		 * @param signedFlag signed flag for the field (when the field is an integral type)
		 */
		void bindParam( int index, enum_field_types type, void* buffer, size_t bufferSize, size_t* valueLength, /*my_*/bool* isNull,
				SignedFlag signedFlag = SignedFlag::UNDEFINED );

		/**
		 * Binds a result to get return values from the query.
		 *
		 * @param index index of the result (field index)
		 * @param result output parameter for the result
		 */
		void bindResult( int index, QueryResult& result );

		/**
		 * Gets the MySQL binding structure to pass the parameters to the prepared statement
		 *
		 * @return the MySQL binding for parameters
		 */
		inline MYSQL_BIND* getQueryParams() {
			return queryParams;
		}

		/**
		 * Gets the MySQL binding structure to get the results from a prepared statement
		 *
		 * @return the MySQL binding for the results
		 */
		inline MYSQL_BIND* getQueryResults() {
			return queryResults;
		}

	private:
		/** the number of parameter fields in the query */
		int numParams;
		/** the number of result fields in the query */
		int numResults;
		/** Parameters to bind */
		MYSQL_BIND* queryParams;
		/** Results to bind */
		MYSQL_BIND* queryResults;
};


/**
 * An SQL prepared statement
 */
class SqlStatement {
		friend class SqlConnection;

	public:

		/**
		 * Destroys the SQL statement.
		 * Releases the resources used by the statement.
		 */
		~SqlStatement();

		/**
		 * Returns the error message if executing the query failed.
		 */
		inline const std::string& getError() const {
			return errorMsg;
		}

		/**
		 * Gets the error number when an error occurred while executing a query
		 * @return the MySQL error number of the last failed operation
		 */
		inline unsigned int getErrno() const {
			return mysqlErrno;
		}

		/**
		 * Executes a SELECT query and stores the results
		 *
		 * @param binding binding of the query parameters and the result fields
		 *
		 * @throw SqlException if an error occurs
		 */
		void executeAndStoreEx( QueryBindings& binding ) noexcept ( false );

		/**
		 * Executes an INSERT or UPDATE query
		 *
		 * @param binding binding of the query parameters and the result fields
		 *
		 * @throw SqlException if an error occurs
		 *
		 * @return the count of affected records
		 */
		my_ulonglong executeUpdateEx( QueryBindings& binding ) noexcept ( false );

		/**
		 * Fetches the next record from the result of a SELECT query
		 *
		 * @return true if and only the next row was successfully fetched.
		 */
		inline bool next() {
			return 0 == mysql_stmt_fetch( sqlStatement );
		}

	protected:

		/**
		 * Constructs a new prepared SQL statement to execute a query.
		 * Used by SqlConnection::prepareStatement();
		 *
		 * @param statement the prepared statement.
		 * @param conMutex the connection's mutex for locking simultaneous queries in concurrent threads
		 */
		SqlStatement( MYSQL_STMT* statement, std::recursive_mutex& conMutex );

		/*
		 * Disable the
		 * - default constructor
		 * - copy constructor
		 * - move constructor
		 * - copy assignment
		 * - move assignment
		 */
		SqlStatement() = delete;
		SqlStatement( const SqlStatement& orig ) = delete;
		SqlStatement( SqlStatement&& ) = delete;
		SqlStatement& operator=( const SqlStatement& ) = delete;
		SqlStatement& operator=( SqlStatement&& ) = delete;

	private:
		/// The prepared statement
		MYSQL_STMT* sqlStatement;
		/// true indicates the result of a SELECT query was stored
		bool resultStored;
		/// error message for the last failed operation
		std::string errorMsg;
		/// MySQL error number of last failed operation
		unsigned int mysqlErrno = 0;
		/// A reference to the connection's mutex to prevent running simultaneous queries in concurrent threads
		std::recursive_mutex& conMutex; // this is intentionally a reference
};

} // namespace db
} // namespace erad

#endif /* IMUTILS_INCLUDE_IMUTILS_EPSQL_H_ */
