/*
 * RefcountDB.h
 *
 *  Created on: Mar 28, 2022
 *      Author: medsrv
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_STUDYDB_H_
#define IMUTILS_INCLUDE_IMUTILS_STUDYDB_H_

#include <string>
#include <list>

#include "imutils/DBConnector.h"

namespace erad::db {


class StudyInfo {
	public:
		std::string study;
		std::string patientName;
		std::string accessionNumber;
		std::string modality;
		std::string patientID;
		std::string studyDate;
		std::string institution;
		std::string studyDescription;
		std::string bodypart;
};


class MPPSInfo {
	public:
		std::string study;
		std::string scheduledStep;
		std::string performedStep;

		MPPSInfo( const std::string& study, const std::string& scheduledStep, const std::string& performedStep )
				: 	study( study ),
					scheduledStep( scheduledStep ),
					performedStep( performedStep ) {
		}
		MPPSInfo( const MPPSInfo& orig ) = delete;
		MPPSInfo( MPPSInfo&& orig ) = default;
		~MPPSInfo() {
		}

		MPPSInfo& operator=( const MPPSInfo& ) = delete;
		MPPSInfo& operator=( MPPSInfo&& ) = delete;
};


class UIDInfo {
	public:
		std::string studyUid;
		std::string proposedUid;
		bool isOrder;

		UIDInfo( const std::string& studyUid, const std::string& proposedUid, bool isOrder )
				: 	studyUid( studyUid ),
					proposedUid( proposedUid ),
					isOrder( isOrder ) {
		}
		UIDInfo( const UIDInfo& orig ) = delete;
		UIDInfo( UIDInfo&& orig ) = default;
		~UIDInfo() {
		}

		UIDInfo& operator=( const UIDInfo& ) = delete;
		UIDInfo& operator=( UIDInfo&& ) = delete;
};


#if 0
// from IMTypes.java - these should be in sync!
/** frozen (unprocessed) - only the original DICOM objects exist */
FROZEN( 0, "STUDY_PREPARINGING_STATUS_FROZEN" ),
/** thawing (processing) - generating processed DICOM objects */
THAWING( 5, "STUDY_PREPARINGING_STATUS_THAWING" ),
/** cold (processed) - processed DICOM objects exist, but there is no cache */
COLD( 10, "STUDY_PREPARINGING_STATUS_COLD" ),
/** cooking (creating cache) - processed DICOM objects exist, the cache is being created */
COOKING( 15, "STUDY_PREPARINGING_STATUS_COOKING" ),
/** partial (created cache) - processed DICOM objects and cache exists, but additional objects are queued */
PARTIAL( 16, "STUDY_PREPARINGING_STATUS_PARTIAL" ),
/** issues (created cache) - processed DICOM objects and cache exists, but some task(s) failed */
ISSUES( 17, "STUDY_PREPARINGING_STATUS_ISSUES" ),
/** finalizing (creating cache) - processed DICOM objects and cache exists, but but global finalization is pending */
FINALIZING( 18, "STUDY_PREPARINGING_STATUS_FINALIZING" ),
/** cooked (full) - both processed DICOM objects and cache exists for the study */
COOKED( 20, "STUDY_PREPARINGING_STATUS_COOKED" );
#endif // 0


enum PrepStat {
	UNKNOWN = -1,
	FROZEN = 0,
	THAWING = 5,
	COLD = 10,
	COOKING = 15,
	PARTIAL = 16,
	ISSUES = 17,
	FINALIZING = 18,
	COOKED = 20
};

/**
 * Interface to get/update information in the study table
 */
class StudyDB {

	public:
		virtual ~StudyDB() {
		}

		/**
		 * Gets the priority of a study.
		 * Checks the priority of the study and - if there is an order with a matching proposed study uid - the order.
		 *
		 * @param study the study instance UID
		 * @param priority output parameter for the priority
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return true if and only the study exists in the study table (either as a study or a matching order)
		 */
		virtual bool getPriority( const std::string& study, int& priority /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets the main status of a study.
		 *
		 * @param study the study instance UID
		 * @param status output parameter for the study status
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return true if and only the study exists in the study table
		 */
		virtual bool getStatus( const std::string& study, int& status /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets study information for logging.
		 * Reads demography and study meta information from the database
		 *
		 * @param study the study instance iud
		 * @param info output parameter for the study information
		 * @throw DBException if an DB error occurs
		 */
		virtual void getStudyLogInfo( const std::string& study, StudyInfo& info /*out*/) noexcept ( false ) = 0;

		virtual void getShortcuts( const std::string& study, std::list<std::string>& shortcuts /*out*/) noexcept ( false ) = 0;

		virtual bool getMinSopI( const std::string& study, std::string& minsopi /*out*/) noexcept ( false ) = 0;

		virtual void getReportUids( const std::string& study, std::list<std::string>& uids /*out*/) noexcept ( false ) = 0;

		virtual bool getReportByInterpretation( const std::string& study, const std::string& interpretation,
				std::string& sopiuid /*out*/) noexcept ( false ) = 0;

		virtual void getRecentStudiesForPatient( const std::string& pid, int days, std::list<std::string>& studies /*out*/)
				noexcept ( false ) = 0;

		virtual bool getProposedUidForOrder( const std::string& order, std::string& proposedUid /*out*/) noexcept ( false ) = 0;

		virtual void getUidInfo( const std::string& study, std::list<UIDInfo>& uids /*out*/) noexcept ( false ) = 0;

		virtual bool getStudyRecord( const std::string& study, bool includeReport, std::vector<SqlField>& fields /*out*/)
				noexcept ( false ) = 0;

		/**
		 * @param mpps  SOP Instance UID of the Modality Performed Procedure Step
		 * @param study Study Instance UID
		 * @param schStep Scheduled Procedure Step ID
		 * @param perfStep Performed Procedure Step ID
		 * @return
		 */
		virtual size_t insertToMpps( const std::string& mpps, const std::string& study, const std::string& schStep,
				const std::string& perfStep ) noexcept ( false ) = 0;

		virtual bool studyExists( const std::string& study ) noexcept ( false ) = 0;

		virtual void getStudiesByScheduledStep( const std::string& schStep, std::vector<std::string>& studies /* out */)
				noexcept ( false ) = 0;

		virtual void getMppsInfo( const std::string& mpps, std::vector<MPPSInfo>& records /*out*/)
				noexcept ( false ) = 0;

		virtual size_t deleteFromMpps( const std::string& mpps, const std::string& study, const std::string& scheduledStep )
				noexcept ( false ) = 0;

		virtual bool getStudyForUid( const std::string& uid, std::string& study /*out*/) noexcept ( false ) = 0;

		virtual bool getPreparingState( const std::string& study, PrepStat &prepStat  /* out*/) noexcept ( false ) = 0;

		virtual bool getNumberOfObjects( const std::string& study, int& numObj ) noexcept ( false ) = 0;

};


/**
 * StudyDB implementation for MySQL
 */
class StudyDBSql : public StudyDB, public PSCache {

	private:

		PsQuery getPriorityQuery;
		PsQuery getStatusQuery;
		PsQuery logInfoQuery;
		PsQuery shortcutsQuery;
		PsQuery minsopiQuery;
		PsQuery reportsQuery;
		PsQuery reportByInterpretationQuery;
		PsQuery studiesForPatientQuery;
		PsQuery proposedUidForOrderQuery;
		PsQuery uidInfoQuery;
		PsQuery mppsInsertQuery;
		PsQuery studyExistsQuery;
		PsQuery studyByScheduledStepQuery;
		PsQuery mppsSelectQuery;
		PsQuery mppsDeleteQuery;
		PsQuery studyForUidQuery;
		PsQuery getPreparingStatusQuery;
		PsQuery getNumObjectsQuery;

	public:

		/**
		 * Constructs a new StudyDBSql object to prepare and execute MySQL queries
		 *
		 * @param pDB connector to the MySQL database
		 */
		StudyDBSql( DBConnectorSql& pDB );
		virtual ~StudyDBSql();

		/**
		 * @copydoc StudyDB::getPriority(const std::string&,int&)
		 */
		virtual bool getPriority( const std::string& study, int& priority /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc StudyDB::getStatus(const std::string&,int&)
		 */
		virtual bool getStatus( const std::string& study, int& status /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc StudyDB::getStudyLogInfo(const std::string&,StudyInfo&)
		 */
		virtual void getStudyLogInfo( const std::string& study, StudyInfo& info /*out*/) noexcept ( false ) override;

		virtual void getShortcuts( const std::string& study, std::list<std::string>& shortcuts /*out*/) noexcept ( false ) override;

		virtual bool getMinSopI( const std::string& study, std::string& minsopi /*out*/) noexcept ( false ) override;

		virtual void getReportUids( const std::string& study, std::list<std::string>& uids /*out*/) noexcept ( false ) override;

		virtual bool getReportByInterpretation( const std::string& study, const std::string& interpretation,
				std::string& sopiuid /*out*/) noexcept ( false ) override;

		virtual void getRecentStudiesForPatient( const std::string& pid, int days, std::list<std::string>& studies /*out*/)
				noexcept ( false ) override;

		virtual bool getProposedUidForOrder( const std::string& order, std::string& proposedUid /*out*/) noexcept ( false ) override;

		virtual void getUidInfo( const std::string& study, std::list<UIDInfo>& uids /*out*/) noexcept ( false ) override;

		virtual bool getStudyRecord( const std::string& study, bool includeReport, std::vector<SqlField>& fields /*out*/)
				noexcept ( false ) override;

		virtual size_t insertToMpps( const std::string& mpps, const std::string& study, const std::string& schStep,
				const std::string& perfStep ) noexcept ( false ) override;

		virtual bool studyExists( const std::string& study ) noexcept ( false ) override;

		virtual void getStudiesByScheduledStep( const std::string& schStep, std::vector<std::string>& studies /* out */)
				noexcept ( false ) override;

		virtual void getMppsInfo( const std::string& mpps, std::vector<MPPSInfo>& records /*out*/)
				noexcept ( false ) override;

		virtual size_t deleteFromMpps( const std::string& mpps, const std::string& study, const std::string& scheduledStep )
				noexcept ( false ) override;

		virtual bool getStudyForUid( const std::string& uid, std::string& study /*out*/) noexcept ( false ) override;

		virtual bool getPreparingState( const std::string& study, PrepStat &procStat  /* out*/) noexcept ( false ) override;

		virtual bool getNumberOfObjects( const std::string& study, int& objNum ) noexcept ( false ) override;
};

} // namespace

#endif /* IMUTILS_INCLUDE_IMUTILS_STUDYDB_H_ */
