/*
 * Caching some object (item) for threads
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_THREADCACHE_H_
#define IMUTILS_INCLUDE_IMUTILS_THREADCACHE_H_


#include <map>
#include <memory>
#include <chrono>
#include <thread>
#include <mutex>
#include <shared_mutex>
#include <functional>


namespace erad::imutils {

// forward declaration, see below
template<class T> class ThreadItem;

using Clock = std::chrono::high_resolution_clock;
using TimePoint = std::chrono::time_point<Clock>;


/**
 * Provides a cache to store some item for threads.
 * Stores a single element for each thread in the cache.
 *
 * Drops the oldest used items (threads) from the cache
 *
 * @tparam T the type of the item to be stored in the cache
 */
template<class T> class ThreadCache {
	public:

		/**
		 * Constructor to create a new cache.
		 *
		 * @param maxSize
		 *              maximum number of elements (threads)
		 * @param creator
		 *              a function to create a new item for the current thread
		 */
		ThreadCache( size_t maxSize, std::function<T* ()>&& creator )
				: 	maxSize( maxSize ),
					creator( creator ) {
		}

		~ThreadCache() {
		}

		/* Copy and move constructors are disabled */

		ThreadCache( const ThreadCache& orig ) = delete;
		ThreadCache( ThreadCache&& orig ) = delete;

		/* Copy and move assignment operators are also disabled */

		ThreadCache& operator=( const ThreadCache& ) = delete;
		ThreadCache& operator=( ThreadCache&& ) = delete;


		/**
		 * Gets the item stored for the current thread.
		 * If no item stored for the current thread, then a new item will be created.
		 * If the cache is full, then the oldest thread will be removed.
		 *
		 * @note It may happen that no old thread can be dropped, because all threads are currently
		 * in use, that is touched while searching the oldest. In this case the actual cache size may
		 * temporarily exceed the max size. At idle time the cache size will drop back to max.
		 *
		 * @warning The returned value should be kept and passed to other methods to keep the
		 * item alive even if gets dropped from the cache meanwhile.
		 * For example getting a sub-element from the item and passing only it without the entire item may
		 * result in a crash.
		 *
		 * @return the item stored or created for the current thread,
		 */
		std::shared_ptr<T> getItem() {
			std::thread::id tid = std::this_thread::get_id();

			try {
				std::shared_lock rlock(cacheMutex);
				return threadCache.at( tid ).getItem();
			}
			catch ( const std::out_of_range& ex ) {
				std::unique_lock wlock( cacheMutex );
				try {
					// Maybe another thread just created it
					return threadCache.at( tid ).getItem();
				}
				catch ( const std::out_of_range& ex ) {
				}

				while ( threadCache.size() >= maxSize ) {
					std::thread::id oldest;
					TimePoint t = std::chrono::system_clock::now();

					for ( auto it = threadCache.begin(); it != threadCache.end(); ++it ) {
						if ( it->second.getLastUsed() < t ) {
							t = it->second.getLastUsed();
							oldest = it->first;
						}
					}

					if ( 0 == threadCache.erase( oldest ) ) {
						// no element can be erased, everything is in use
						break;
					}
				}
				std::shared_ptr<T> sp( creator() );
				threadCache.emplace( tid, sp );
				return sp;
			}
		}

		/**
		 * Clears the cache.
		 */
		void clear() {
			std::unique_lock wlock( cacheMutex );
			threadCache.clear();
		}


	private:

		/// The cache
		std::map<std::thread::id, ThreadItem<T>> threadCache;
		/// Maximum size for the cache
		size_t maxSize;
		/// mutex for adding new elements (and dropping old ones) */
		mutable std::shared_mutex cacheMutex;;

		/// A function to create a new element for the current thread
		std::function<T* ()> creator;
};


/**
 * A helper class to store an item in the @code{ThreadCache}.
 * Stores the item and maintains the last used time (last requested time) of the item.
 *
 * @tparam T the type of the item to be stored
 */
template<class T> class ThreadItem {
		friend class ThreadCache<T> ;

	public:

		/**
		 * Constructor to create a new ThreadItem to store the particular item.
		 *
		 * @param item
		 *              the item to store
		 */
		ThreadItem( std::shared_ptr<T> item )
				: item( item ) {
			lastUsed = std::chrono::system_clock::now();
		}

		/**
		 * Destroys @code{this} ThreadItem
		 */
		~ThreadItem() {
		}

	private:
		/* Copy and move constructors are disabled */

		ThreadItem( const ThreadItem& orig ) = delete;
		ThreadItem( ThreadItem&& orig ) = delete;

		/* Copy and move assignment operators are also disabled */

		ThreadItem& operator=( const ThreadItem& ) = delete;
		ThreadItem& operator=( ThreadItem&& ) = delete;


		/**
		 * Gets the last used (last requested) time of the item
		 * @return the last used (last requested) time of the item
		 */
		inline const TimePoint& getLastUsed() const {
			return lastUsed;
		}

		/**
		 * Gets the item.
		 * By returning a shared pointer the item remain alive even if removed from the cache while still being used.
		 *
		 * @return the item.
		 */
		inline std::shared_ptr<T> getItem() {
			lastUsed = std::chrono::system_clock::now();
			return item;
		}

	private:
		/// The item to store
		std::shared_ptr<T> item;

		/// Last used time of the item;
		TimePoint lastUsed;
};

} // erad::imutils

#endif /* IMUTILS_INCLUDE_IMUTILS_THREADCACHE_H_ */
