/*
 * pbdcm/dtregistry.h
 *
 *  Created on: Dec 13, 2013
 *      Author: tnemeth
 */

#ifndef DTREGISTRY_H_
#define DTREGISTRY_H_

namespace pbtz {


/**
 * Defines tag pairs for date/time values stored in separate DA and TM elements in DICOM objects.
 */
class DateTimeTagPair {
	protected:
	/** Tag of the date element */
	const DcmTagKey m_date;
	/** Tag of the time element */
	const DcmTagKey m_time;

	public:

	/**
	 * Constructor creating a new DateTimeTagPair object
	 *
	 * @param date   Tag of the date element
	 * @param time   Tag of the time element
	 */
	DateTimeTagPair( const DcmTagKey & date, const DcmTagKey & time ) : m_date( date ), m_time( time ) {
	}

	/**
	 * Destructor destroying this DateTimeTagPair object
	 */
	virtual ~DateTimeTagPair() {}

	/**
	 * @return Tag of the date element
	 */
	inline const DcmTagKey & getDate() const {
		return m_date;
	}

	/**
	 * @return Tag of the time element
	 */
	inline const DcmTagKey & getTime() const {
		return m_time;
	}

	/**
	 * Checks whether a DICOM tag matches to the time tag of this DateTimeTagPair object
	 * @param tag   the DICOM tag to check
	 * @return true if and only if the argument <code>tag</code> has the same tag key as the time element of this DateTimeTagPair object
	 */
	inline bool isTimeTagMatches( const DcmTag & tag ) const {
		return tag == m_time;
	}

};


/**
 * Defines tag pairs for date/time values stored in separate DA and TM private elements in DICOM objects.
 */
class PrivateDateTimeTagPair : public DateTimeTagPair {
	/** Private creator code */
	OFString m_privateCreator;

	public:

	/**
	 * Constructor creating a new PrivateDateTimeTagPair object
	 *
	 * @param date             Tag of the date element
	 * @param time             Tag of the time element
	 * @param privateCreator   Private creator code
	 */
	PrivateDateTimeTagPair( const DcmTagKey & date, const DcmTagKey & time, const char * privateCreator )
	: DateTimeTagPair( date, time ) {
		m_privateCreator = privateCreator;
	}

	/**
	 * Destructor destroying this PrivateDateTimeTagPair object
	 */
	virtual ~PrivateDateTimeTagPair() {}

	/**
	 * @return the private creator code
	 */
	inline const OFString & getPrivateCreator() const {
		return m_privateCreator;
	}

	/**
	 * Checks whether a DICOM tag matches to the time tag of this DateTimeTagPair object.
	 *
	 * An argument time tag matches to this PrivateDateTimeTagPair if all of the following criteria applies:
	 * - the group key of the argument tag and the time tag of this PrivateDateTimeTagPair object is the same
	 * - the time tag of this object resides in the provided private block
	 * - the element key of the argument tag is the same as the time tag of this PrivateDateTimeTagPair within the private block
	 *
	 * @param tag           the DICOM tag to check
	 * @param privateBlock  the private block the tag resides in
	 * @return true if and only if the argument <code>tag</code> has the same tag key as the time element of this DateTimeTagPair object
	 */
	inline bool isTimeTagMatches( const DcmTag & tag, Uint16 privateBlock ) const {
		return ( tag.getGroup() == m_time.getGroup() )
				&& ( ( tag.getElement() & 0xff00 ) == privateBlock )
				&& ( ( tag.getElement() & 0x00ff ) == m_time.getElement() );
	}

};

/**
 * Registry of Date/Time tag pairs supported in timezone conversions.
 */
class DateTimeRegistry {
	public:

		/**
		 * Constructor creating a new DateTimeRegistry object
		 */
		DateTimeRegistry() {}

		/**
		 * Loads additional Date/Time tag pairs
		 *
		 * This method does nothing, but is prepared to support loading date/time key pairs from a file at a later time for sites
		 * using date/time fields not defined by the DICOM standard
		 */
		void load() {}

		/**
		 * Searches the registry for a DateTimeTagPair object for a given dateTag
		 *
		 * @param dateTag       the date tag to find a DateTimeTagPair object for
		 * @param privateGroup  the private block for private Date elements
		 *
		 * @return a pointer to the DateTimeTagPair object found or NULL if date was not found in the registry
		 */
		const DateTimeTagPair * findForDate( const DcmTag & dateTag, Uint16 & privateGroup ) const;
};

}

#endif /* DTREGISTRY_H_ */
