/*
 * convrules.h
 *
 *  Created on: Jan 12, 2015
 *      Author: tnemeth
 */

#ifndef CONVRULES_H_
#define CONVRULES_H_

#include <dcmtk/ofstd/ofstd.h>
#include <dcmtk/dcmdata/dcxfer.h>

namespace imconvertts {

/**
 * Identifies built in rule sets
 */
typedef enum {
	// Always - enable any conversion
	RULE_SET_ALWAYS = 0,

	// If not in any supported (archive) compression format
	RULE_SET_ARCHIVE = 1,

	// Only if uncompressed
	RULE_SET_UNCOMPRESSED = 2,

// To support custom rule set read from a configuration file
// RULE_SET_CUSTOM = 3,
} RULE_SET_TYPE;


/**
 * Registry of Transfer Syntax Aliases
 * Conversion rule sets can refer to Individual transfer syntaxes or Transfer Syntax Aliases defined here
 */
class TSAliasRegistry {
	private:

		/** The registry of Transfer Syntax Aliases : alias -> TS list */
		OFMap< OFString, OFList< E_TransferSyntax > > m_registry;

	public:

		/**
		 * Constructor to create a new Transfer Syntax Alias registry.
		 * It also sets up built in Transfer Syntax Aliases
		 */
		TSAliasRegistry();

		/**
		 * Prints the Transfer Syntax Aliases to the standard output
		 */
		void print() const;

		/**
		 * Checks whether the provided Transfer Syntax is matching to the provided name.
		 * The name can be either the UID or name of an individual Transfer Syntax (see XferNames in dcmtk/dcmdata/dcxfer.cc),
		 * or the name of a Transfer Syntax Alias
		 *
		 * @param name  name of the TS or TS Alias
		 * @param ts    The Transfer Syntax to check
		 *
		 * return true if and only if the Transfer Syntax matches to the specified name (group or individual TS)
		 */
		bool isMatching( const OFString & name, E_TransferSyntax ts ) const;
};


/**
 * Describes an individual rule to be used in a rule set.
 *
 * A rule defines whether the conversion between two compression formats is allowed or not.
 * Both 'from' and 'to' Compression formats can either be the name or UID of an individual Transfer Syntax
 * or the name of a Transfer Syntax alias.
 *
 * See TSAliasRegistry for Transfer Syntax alias names
 */
class ConvRule {
		friend class RuleSet;

	private:

		/** The from compression format */
		OFString m_from;
		/** The to compression format */
		OFString m_to;

		/** Indicates whether conversion from 'from' to 'to' is allowed or not */
		bool m_isAllowed;

	protected:

		/**
		 * Constructor to create a new ConvRule object
		 *
		 * @param from        The from compression format
		 * @param to          The to compression format
		 * @param isAllowed   Indicates whether conversion from 'from' to 'to' is allowed or not
		 */
		ConvRule( const OFString & from, const OFString & to, bool isAllowed );

		/**
		 * @return the 'from' compression format
		 */
		inline const OFString & getFrom() const {
			return m_from;
		}

		/**
		 * @return the 'to' compression format
		 */
		inline const OFString & getTo() const {
			return m_to;
		}

		/**
		 * @return true if and only if conversion from 'from' to 'to' is allowed by this rule
		 */
		inline bool isAllowed() const {
			return m_isAllowed;
		}
};


/**
 * Set of rules to determine whether a conversion from one Transfer Syntax to another is allowed or not.
 *
 * The first rule matching to both the 'from' and 'to' Transfer Syntaxes shall be used. If there is no matching rule,
 * then the conversion shall be allowed.
 */
class RuleSet {

		/** The conversion rules in apply order */
		OFList< ConvRule > m_rules;

		/** Rule set information*/
		OFString m_info;

	protected:
		/**
		 * Initializes the default rule set allowing conversion from ANY Transfer Syntax to ANY Transfer Syntax
		 */
		void initAlwaysRuleSet();
		/**
		 * Initializes the 'Archive' rule set allowing conversion from non-archive Transfer Syntaxes to ANY Transfer Syntaxes
		 */
		void initArchiveRuleSet();
		/**
		 * Initializes the 'Uncompressed' rule set allowing only conversion from Uncompressed Transfer Syntaxes
		 */
		void initUncompressedRuleSet();

		/**
		 * Clears the current rule set
		 */
		inline void clear() {
			m_rules.clear();
		}

	public:
		/**
		 * Constructor to create a new RuleSet object.
		 * The rule set shall be initialized according to the ruleSetType parameter (see RULE_SET_TYPE)
		 *
		 * @param ruleSetType  identifies the rule set to use
		 */
		RuleSet( RULE_SET_TYPE ruleSetType );

		/**
		 * Checks whether or not conversion from 'tsFrom' Transfer Syntax to 'tsTo' Transfer Syntax is allowed by this rule set.
		 * @param tsFrom   from Transfer Syntax
		 * @param tsTo     to Transfer Syntax
		 * @return true if and only if conversion from 'tsFrom' Transfer Syntax to 'tsTo' Transfer Syntax is allowed by this rule set
		 */
		bool isAllowed( E_TransferSyntax tsFrom, E_TransferSyntax tsTo ) const;

		/**
		 * Modifies the rule set to use.
		 */
		void set( RULE_SET_TYPE newRuleSet );

		/**
		 * @return rule set information
		 */
		inline const OFString & getInfo() const {
			return m_info;
		}
};

/**
 * Return the name of a rule set type
 */
const char * getRuleSetType( RULE_SET_TYPE ruleSetType );

/**
 * For testing
 */
bool testConvRules();

} /* namespace imconvrules */

#endif /* CONVRULES_H_ */
