#!/bin/bash

#
#
#  batchretrieve.sh
#
#  tool to get all files based on a certain date range from a DICOM Q/R device.
#
#  Usage:
#  ./batchretrieve.sh (without options) -- interactive mode. 
#
#  In interactive mode, batchretrieve.sh will prompt for the parameters,
#  such as the IP, port, AETITLE, QR model, local AE title, and the date range.
#  If run in interactive mode, it will also create a config file,
#  in ~/var/conf/batchretrieve.rc for defaults
#
#  In non-interactive mode, the parameters are as follows:
#
#  -i Q/R IP address
#  -p Q/R port
#  -a Q/R AETITLE
#  -r Q/R model (root)
#  -t local AETITLE
#  -d date range
#  -l logfile 
#
#


. ~/etc/setenv.sh

. $DICOM_VAR/pb-scp.cfg
[ -z "$AE_TITLE" ] && AE_TITLE="$DICOM_AE_TITLE"

OK=""

if [ $# = 0 ] ; then

 while [ -z "$OK" ] ; do

  [ -f $CONF/batchretreive.rc ] && . $CONF/batchretreive.rc
  
  echo -n "Q/R IP address [$QRIP]:"
  read QRIP1
  [ -n "$QRIP1" ] && QRIP=$QRIP1

  echo -n "Q/R port [$QRPORT]:"
  read QRPORT1
  [ -n "$QRPORT1" ] && QRPORT=$QRPORT1

  echo -n "Q/R AETITLE [$QRAE_TITLE]:"
  read QRAE_TITLE1
  [ -n "$QRAE_TITLE1" ] && QRAE_TITLE=$QRAE_TITLE1

  QRROK=""
  
  while [ -z "$QRROK" ] ; do
    echo -n "Q/R model ([P]atient/[S]tudy/PatientStudy[O]nly P/S/O) [$QRROOT]:"
    read QRROOT1
    [ -n "$QRROOT1" ] && QRROOT=$QRROOT1
    if [ "$QRROOT" = "P" -o "$QRROOT" = "S" -o "$QRROOT" = "O" ] ; then QRROK=1; fi
  done

  echo -n "local AETITLE [$AE_TITLE]:"
  read AE_TITLE1
  [ -n "$AE_TITLE1" ] && AE_TITLE=$AE_TITLE1

  echo -n "date range (format: 20030125- or 20030125-20030214) [$DATERANGE]:"
  read DATERANGE1
  [ -n "$DATERANGE1" ] && DATERANGE=$DATERANGE1

  echo -n "Debug (y/n):"
  read QRDEBUG1
  QRDEBUG=" -v "
  [ "$QRDEBUG1" = "y" -o "$QRDEBUG1" = "Y" ] && QRDEBUG=" -v -d "

  LOGFILE="/tmp/br_$DATERANGE.log"
  echo -n "Log file [$LOGFILE]:"
  read LOGFILE1
  [ -n "$LOGFILE1" ] && $LOGFILE=$LOGFILE1
  

  echo "QRIP=\"$QRIP\""              > $CONF/batchretreive.rc
  echo "QRPORT=\"$QRPORT\""         >> $CONF/batchretreive.rc
  echo "QRAE_TITLE=\"$QRAE_TITLE\"" >> $CONF/batchretreive.rc
  echo "QRROOT=\"$QRROOT\""         >> $CONF/batchretreive.rc
  echo "AE_TITLE=\"$AE_TITLE\""     >> $CONF/batchretreive.rc
  echo "DATERANGE=\"$DATERANGE\""   >> $CONF/batchretreive.rc

  echo  
  echo -n "Is it all correct (Y/N)?"
  read r
  if [ "Y" = "$r" -o "y" = "$r" ] ; then OK=1 ; fi
  echo

 done
else

  QRPORT=104
  QRROOT=P
  
  [ -f $CONF/batchretreive.rc ] && . $CONF/batchretreive.rc
  while getopts "i:p:a:r:t:d:l:" OPT ; do
    case $OPT in 
     i) QRIP="$OPTARG" ;;
     p) QRPORT="$OPTARG" ;;
     a) QRAE_TITLE="$OPTARG" ;;
     r) QRROOT="$OPTARG" ;;
     t) AE_TITLE="$OPTARG" ;;
     d) DATERANGE="$OPTARG" ;;
     l) LOGFILE="$OPTARG" ;;
    esac
  done

[ -z "$LOGFILE" ] && LOGFILE="/tmp/br_$DATERANGE.log"

if [ -z "$QRIP" -o -z "$QRAE_TITLE" ] ; then
  echo "No IP or AE_TITLE given" | tee -a $LOGFILE
  exit
fi

#  echo "Q/R IP address = $QRIP"
#  echo "Q/R port = $QRPORT"
#  echo "Q/R AETITLE = $QRAE_TITLE"
#  echo "Q/R root = $QRROOT"
#  echo "Local AETITLE = $AE_TITLE"
#  echo "Date range = $DATERANGE"
#  echo "Log file = $LOGFILE"
  
#  exit

fi

#
#  DCM file for find query
#

  echo "(0008,0052) CS [STUDY]
(0020,000d) UI [] # Study IUID
(0020,1208) IS [] # Number of Study related images
(0010,0010) PN [] # Patient Name
(0008,0020) DA [] # Study Date
(0008,0030) TM [] # Study Time
(0008,0061) CS [] # Modality (-ies)
" > /tmp/brf.dmp
  ~/component/dcmtk/bin/dump2dcm /tmp/brf.dmp /tmp/brf.dcm >/dev/null 2>&1
  
#
# DCM file for move
#  

  echo "(0008,0052) CS [STUDY]" > /tmp/br.dmp
  ~/component/dcmtk/bin/dump2dcm /tmp/br.dmp /tmp/br.dcm >/dev/null 2>&1

#
# Logfile header
#  

  echo `date` "Getting studies in the range of $DATERANGE from $QRIP:$QRPORT/$QRAE_TITLE in '$QRROOT' model" | tee -a $LOGFILE


#
# Do a find and process result to log matching studies
#

  ~/component/dcmtk/bin/findscu $QRDEBUG -$QRROOT --aetitle $AE_TITLE \
       --key 0008,0020="${DATERANGE}" --key 0008,0030="000000-" \
       --call $QRAE_TITLE $QRIP $QRPORT /tmp/brf.dcm 2>/tmp/brfinderr.log >/tmp/brfind.log 

  if [ "$?" != 0 ] ; then
    echo "Error querying archive" | tee -a $LOGFILE
    cat /tmp/brfinderr.log | tee -a $LOGFILE
    exit
  fi

>/tmp/brstyiuid.log
       
cat /tmp/brfind.log | awk '
function getdata()
{
  m=index($0,"[");
  if (m!=0)
  {
    r=substr($0,m+1);
    return substr(r,1,index(r,"]")-1)
  }
  else return "-";
}
$1=="RESPONSE:"   { responses=1;}
$1=="(0008,0020)" { date=getdata(); } 
$1=="(0008,0030)" { time=getdata(); }
$1=="(0008,0061)" { mod=getdata(); }
$1=="(0010,0010)" { pname=getdata(); }
$1=="(0020,000d)" { styiuid=getdata(); if (responses==1) {print pname,mod,date,time, styiuid; print styiuid >>"/tmp/brstyiuid.log"; } }
$1=="(0020,1208)" { a=getdata(); if (responses==1) { obj+=a; stn+=1 } } 
END { print "Total number of studies (objects) to be transferred: " stn " (" obj ")" ; } 
' | tee -a  $LOGFILE

rm /tmp/brfind.log


#
#  Do the actual move
#

ERROR=""

cat /tmp/brstyiuid.log | while read styiuid; do
  echo -n Moving $styiuid... | tee -a $LOGFILE
  
  ~/component/dcmtk/bin/movescu $QRDEBUG -$QRROOT --move $AE_TITLE --aetitle $AE_TITLE \
       --key 0020,000d=$styiuid \
       --call $QRAE_TITLE $QRIP $QRPORT /tmp/br.dcm 2>/tmp/brmoverr.log >/tmp/brmove.log

  if [ "$?" != 0 ] ; then
    echo "Error moving $styiuid from archive" | tee -a $LOGFILE
    cat /tmp/brmoverr.log | tee -a $LOGFILE
    ERROR=1
  fi
  echo "  OK" | tee -a $LOGFILE
  
done


  if [ -z "$ERROR" ] ; then
    echo "Done: Successfully moved all studies" | tee -a $LOGFILE
  else
    echo "Done: some errors encountered, see logfile" | tee -a $LOGFILE    
  fi
  
  echo >> $LOGFILE
  echo >> $LOGFILE
  echo >> $LOGFILE
