#!/bin/bash
#
# Component: tools
#
# Description:
#    Prints queue status information, based on THIS server's database
#
# Usage:
#      qst.sh [-h] [-P] [-H] [unprocessed|scheduled|retry|failed|suspended|delay|future]
#
#

BaseName=$( basename $(readlink -f "$0" ) )


# Controls number formatting in printf (default is no thousand separators)
NumberFormat="%15d"
# The queue to display
Queue=""
# Print in a processable output format
Processable="false"
# count tasks only with priority under this - default 0 means all tasks
PriUnder="0"

#
# Prints a line to the task count report
# (indented)
#
# $1 queue
# $2 number of tasks
#
function printLine_() {
	local queue="$1"
	local cnt="$2"

	printf "%4s%-15s ${NumberFormat}\n" " " "${queue}:" "${cnt}"
}


#
# Prints a line with total task counts
# (unindented)
#
# $1 label
# $2 number of tasks
#
function printTotalLine_() {
	local label="$1"
	local cnt="$2"

	printf "%-19s ${NumberFormat}\n" "${label}:" "${cnt}"
}


#
# Prints a line to a detailed task list
# (indented)
#
# $1 queue
# $2 number of tasks
#
function printDetailedLine_() {
	local queue="$1"
	local cnt="$2"

	printf "%4s%-45s ${NumberFormat}\n" " " "${queue}:" "${cnt}"
}


#
# Prints a line with total task counts for a detailed list
# (unindented)
#
# $1 label
# $2 number of tasks
#
function printDetailedTotalLine_() {
	local label="$1"
	local cnt="$2"

	printf "%-49s ${NumberFormat}\n" "${label}:" "${cnt}"
}



#
# Prints usage information
# [$@] - error message
#
function usage_() { 

	cat <<-EOF
		${BaseName} [-h] [-H] [-P [ -u <priority> ] ] [<queue>]

		       -h              : prints this help
		       -H              : display task counts in a human readable format
		                         (grouped with thousands according to locale settings)
		       -P              : display output in a processable format
		                         prints only total tasks in the queue
		                         ignores -H
		           -u priority : usable with -P only! count tasks under the given priority only          
		       <queue>         : display tasks within the particular queue. Supported queues:
		                            - unprocessed
		                            - scheduled
		                            - retry
		                            - failed
		                            - suspended
		                            - delay
		                            - future

		Prints the number of tasks grouped by task types in the queue provided on the command line, or
		prints the number of tasks in each queue when no queue provided. 
		Allways based on THIS server's database
	EOF

	if [ -n "$1" ] ; then
		echo >&2
		echo "$@" >&2
		exit 1
	fi

	exit 0
}


while getopts ":hHPa:u:l" OPT ; do
	case $OPT in
		h)
			usage_
			;;
		P)
			Processable="true"
			;;
		H)
			NumberFormat="%'15d"
			;;
			
		u)  PriUnder=$OPTARG ;;		
		
		*)
			usage_ "Unrecognized option: '-$OPTARG'"
			;;
	esac
done


shift $(( OPTIND - 1 ))
Queue="$1"


#
# Print task counts for each queue + totals
#
if [ -z "${Queue}" ] ; then
	echo "Tasks in files:"
	num=0
	actnum=$( find $HOME/var/taskqueue/ -maxdepth 1 -type f | wc -l )
	printLine_ "unprocessed" "$actnum"

	num=$(($actnum+$num))
	for qq in $(find $HOME/var/taskqueue/ -mindepth 1 -maxdepth 1 -type d | sort | awk -F / '{print $6}') ; do
		actnum=$( find $HOME/var/taskqueue/${qq} -mindepth 1 -maxdepth 1 -type f | wc -l )
		if [ $actnum -ne 0 ] ; then 
			printLine_ "${qq}" "${actnum}"
		fi
		num=$(($actnum+$num))
	done
	echo "============"
	printTotalLine_ "Total" "$num"
	echo

	numd=0
	echo "Tasks in database:"
	for qq in unprocessed delay failed future retry scheduled suspended ; do
		taskCnt=$( $TOOLS_HOME/lsql.sh "select count(*) from TaskQueue where TQQUEUE='$qq'" -N )
		if [ "${qq}" != "unprocessed" ] || [ $taskCnt -ne 0 ] ; then
			printLine_ "${qq}" "${taskCnt}"
		fi
		numd=$(($taskCnt+$numd))
	done
	echo "============"
	printTotalLine_ "Total" "$numd"
	echo
	echo "============"
	printTotalLine_ "Grand Total" $(( $num + $numd ))
	echo
	
	exit 0
fi


#
# Print task counts grouped by taks types in a particular queue
#


if [ "${Queue}" = "running" ] ; then
	. ~/etc/setenv.sh; $TASKD_HOME/runjavacmd -c "info" | awk -F "\t" '
	 (NF != 0) {print $11 "\t" $20 "\t" ($15/1000000000) "_seconds\t" $18}
	' | column -t
else 

	if [ "${Queue}" = "unprocessed" ] ; then
		DIR="$HOME/var/taskqueue"
	else
		DIR="$HOME/var/taskqueue/.${Queue}"
	fi
	
	
	if [ "$Processable" = "true" ] ; then
		# called from code, return simple, processable format
		if [ "$PriUnder" != "0" ] ; then 
			filenum=$( find $DIR -mindepth 1 -maxdepth 1 -type f 2>/dev/null | awk -F / '{print $7}' | awk -F . -v pri=$PriUnder '
				BEGIN {
					n=0;
				}
				{
					if( $1 < pri ) { 
						n++;
					}
				}
				END {
					printf n;
				}
			' )
		
			taskCnt=$( $TOOLS_HOME/lsql.sh "select count(TQQUEUE) from TaskQueue where TQQUEUE = '${Queue}' AND TQPRI<'${PriUnder}';" -N 2>/dev/null )
		else 
			filenum=$( find $DIR -mindepth 1 -maxdepth 1 -type f | wc -l )
			taskCnt=$( $TOOLS_HOME/lsql.sh "select count(TQQUEUE) from TaskQueue where TQQUEUE = '${Queue}';" -N 2>/dev/null )
		fi
		if [ $? = 0 ] ; then
			echo $(( $filenum + $taskCnt ))
		else
			echo $filenum
		fi	
		exit 0
	fi
	
	
	echo
	echo "Tasks in files - Detailed status of the ${Queue} queue per priority:"
	echo
	echo "prty  #tasks"
	echo "------------"
	
	find $DIR -mindepth 1 -maxdepth 1 -type f 2>/dev/null | awk -F / '{print $7}' | awk -F . -v dir=$DIR -vnumFormat="${NumberFormat}" '
		{
			d[$1]++;
			n++;
			s[$1] = $0;
		}
	
		END {
			for ( i in d ) {
				f = dir "/" s[i];
				getline cmd0 < f;
				split( cmd0, cmd1, " " );
				nf = split( cmd1[1], cmd2, "/" );
				ss = cmd2[ nf ];
				printf "%4s  " numFormat "   ( ?? %s ??)\n",i, d[i],ss | "sort";
			}
			close("sort")
			printf "============\n";
	#		printf "%-19s " numFormat "\n", "Total:", n;
		}'
	totalnum=$( find $DIR -mindepth 1 -maxdepth 1 -type f 2>/dev/null| wc -l )      #there seems to be no way to get it easily from awk, and it will be cached anyway
	printDetailedTotalLine_ "Total" "${totalnum}"
	
	
	echo
	echo
	echo "Tasks in database - Detailed status of the ${Queue} queue per type:"
	echo
	echo "type  #tasks"
	echo "------------"
	
	$TOOLS_HOME/lsql.sh "select TQSCRIPTNAME as CMD, count(*) as N from TaskQueue where TQQUEUE = '$Queue' group by CMD" -N | while read taskCmd count ; do
		printDetailedLine_ "${taskCmd}" "${count}"
	done
	
	echo "============"
	taskCnt=$($TOOLS_HOME/lsql.sh "select count(TQQUEUE) from TaskQueue where TQQUEUE = '$Queue';" -N)
	printDetailedTotalLine_ "Total" "${taskCnt}"
	echo
	echo "============"
	printDetailedTotalLine_ "Grand Total" $(( $totalnum + $taskCnt ))
	echo

fi

 
exit 0
