#!/bin/bash
#
# Reschedules suspended tasks
#
# Usage:
#    rescheduleSuspendedTasks.sh [-l <limit>] [-n <num>] [-s <sleep>] [-t <target>] <suspend directory>
#
#         -l <limit>  : start rescheduling tasks if number of tasks in target queue is below this limit [default: 1000]
#         -n <num>    : number of tasks to be rescheduled at a time [default: 3000]
#         -s <sleep>  : time to sleep between checking the queue (sec) [default: 30]
#         -t <target> : OBSOLETE: will be ignored; it was the target queue in the past
#         <suspend directory> : reschedule tasks from this directory
#
# Last Updated     : $Author $
# Update Date Time : $Date $
# RCS File Name    : $RCSfile $
# RCS Revision     : $Revision $
# Symbolic name    : $Name $
# State            : $State $
#

. $HOME/etc/setenv.sh

#
# Prints usage information and exit
#
# [$1] - error message to be printed before usage information
#
function usage_()
{

	if [ -n "$1" ] ; then
		cat <<-EOFERROR >&2

			$1

		EOFERROR
	fi
	
	cat <<-EOFUSAGE
		rescheduleSuspendedTasks.sh [-l <limit>] [-n <num>] [-s <sleep>] <suspend directory>

		    -l <limit>  : start rescheduling tasks if number of tasks in the scheduled queue is below this limit [default: 1000]
		    -n <num>    : number of tasks to be rescheduled at a time [default: 3000]
		    -s <sleep>  : time to sleep between checking the queue (sec) [default: 30]
		    <suspend directory> : reschedule tasks from this directory

	EOFUSAGE

	[ -n "$1" ] && exit 1
	exit 0
}


TaskNumLimit="1000"
RescheduleCount="3000"
SleepSecs="30"
TargetQueue="${TASKD_GLOBALQUEUE}"

while getopts "hl:n:s:t:" OPT ; do
	case $OPT in 
		l) TaskNumLimit=$(( $OPTARG + 0 )) ;;
		n) RescheduleCount=$(( $OPTARG + 0 )) ;;
		s) SleepSecs=$(( $OPTARG + 0 )) ;;
		t) echo "Warning: -t $OPTARG option is obsolete (will be ignored)" ;;
		h) usage_ ;;
		*) usage_ "Unrecognized option: $OPT" ;;
	esac
done

[ 0 -eq $TaskNumLimit ] && usage_ "Invalid argument for task number limit (-l)"
[ 0 -eq $RescheduleCount ] && usage_ "Invalid argument for reschedule count (-n)"
[ 0 -eq $SleepSecs ] && usage_ "Invalid argument for sleep seconds (-s)"

shift $(( $OPTIND - 1 ))

SuspendedQueue="$1"

[ -z "$SuspendedQueue" ] && usage_ "Missing argument: <suspend directory>"
[ ! -d "$SuspendedQueue" ] && usage_ "Invalid suspend directory: $SuspendedQueue"

if [ -z "$TargetQueue" ] ; then
	echo "The TASKD_GLOBALQUEUE is not set in etc/setenv.sh" >&2
	exit 1
fi


cat <<-EOFINFO
	Suspend directory: $SuspendedQueue
	Target queue:      $TargetQueue
	Limit:             $TaskNumLimit
	Reschedule count:  $RescheduleCount
	Sleep seconds:     $SleepSecs
EOFINFO


# Checks the count of the tasks in the TargetQueue queue
#
# Return 0 if the number of tasks in the queue below the $TaskNumLimit
#
checkQueues_()
{
#	taskCnt=`ls -1 $TargetQueue $TargetQueue/.scheduled/ | wc -l`
	taskCnt=$(sql.sh "select count(*) from TaskQueue where TQQUEUE='scheduled'" -N)
	echo "$taskCnt :: $TaskNumLimit"
	[ $taskCnt -lt $TaskNumLimit ] && return 0
	return 1
}


#
# Takes the first study from $StudyListFile and
# reschedule put tasks for its objects
#
#
doRestore_()
{
	CWD=`pwd`
	cd "$SuspendedQueue"
	if [ $? -ne 0 ] ; then
		echo "Could not change current directory to $SuspendedQueue" >&2
		exit 1
	fi

	echo "Rescheduling tasks"
	ls | head -n $RescheduleCount | xargs mv --target-directory=$TargetQueue/
}

# While there are studies to be fixed
while [ -n "`ls $SuspendedQueue`" ] ; do
	if checkQueues_ ; then
		echo "Start rescheduling tasks"
		doRestore_
	fi

	sleep $SleepSecs
done
