#!/bin/bash
#
# Component: tools
#
# Description:
#   Restores archive database from the content of the archive (with -d) or from the insert.sql scripts
#
# Usage:
#   restorearchivedb.sh [-o] [-d] [-q] {-a | -s <studydir>}
#
#		-o	: SQL dump only (Do not execute the REPLACE statements.)
#						(default: executes the REPLACE statements)
#		-d	: it does not use the sql script from the .info directory, it figures out the data
#     -q : quick mode for figuring out Modalitites in study - will only check MIN SOP I object
#          [much more faster but less accurate - may give false value for Modalities if study consist of several modalities]
#		-a	: iterate through all studies in the archive directory and fix them all 
#		-s	: fix only the entry for the specified study directory
#		                the parameter expected to be the archive directory of a study relative to
#		                \$ARCHIVE_DIR. Eg. a1/199908/<study iuid>
#
# Last Updated     : $Author$
# Update Date Time : $Date$
# RCS File Name    : $RCSfile$
# RCS Revision     : $Revision$
# Symbolic name    : $Name$
# State            : $State$
#

. $HOME/etc/setenv.sh

GETDATA_AWK="$TOOLS_HOME/getdata.awk"

usage_() {
	[ -n "$1" ] && echo "$1" >&2
	cat <<-ENDOFUSAGE
		 restorearchivedb.sh [-o] [-d] [-q] {-a | -s <studydir>}

		    -o        : dump only; do not execute SQL queries
		    -d        : it does not use the sql script from the .info directory, it figures out the data
		    -q        : quick mode for figuring out Modalitites in study - will only check MIN SOP I object
                      [much more faster but less accurate;
		                 may give false value for Modalities if study consist of several modalities]
		    -a        : iterate through all studies in the archive directory and fix them all 
		    -s        : fix only the entry for the specified study directory
		                the parameter expected to be the archive directory of a study relative to
		                \$ARCHIVE_DIR. Eg. a1/199908/<study iuid>
	ENDOFUSAGE

	exit 1
}


dumpOnly="false"
fixAll="false"
notUseSQL="false"
quickMode="false"
studydir=""

while getopts 'doqas:' OPT; do
	case $OPT in
	s) studydir="$OPTARG" ;;
	o) dumpOnly="true" ;;
	a) fixAll="true" ;;
	d) notUseSQL="true" ;;
	q) quickMode="true" ;;
	*) usage_ "Invalid Parameter: $OPT" ;;
	esac
done

shift $(( $OPTIND - 1 ))
[ -n "$1" ] && usage_ "Invalid argument: $1"

[ -z "$studydir" -a "$fixAll" != "true" ] && usage_ "Exactly one of -a or -s options should be specified"
[ -n "$studydir" -a "$fixAll" = "true" ] && usage_ "Exactly one of -a or -s options should be specified"

#echo "$dumpOnly  --  $fixAll -- $studydir"
#exit 0


##
#Usage: getDCMvalue <DICOM-TAG> <PATH OF DICOM-FILE>
#
getDCMvalue() {
	_val=`$DICOM_HOME/bin/dcmdump +P "$1" "$2" | sed 's/^.*\[\(.*\)\].*$/\1/'`
	[ $? -ne 0 ] && _val=""
	_isEmpty=`echo "$_val" | grep "(no value available)"`
	[ -n "$_isEmpty" ] && _val=""
	echo "$_val"
	return 0
}

##
# check if a file is in the obsoleted file list of the study
#
# $1 - file to be checked
# $2 - obsoleted file list
#
is_obsoleted_()
{
	[ -z "$2" ] && return 0

	for i in $2 ; do
		[ "$i" = "$1" ] && return 0
	done

	return 1
}

#
# Combines values of DICOM DA and TM tags into a MySQL DATETIME value
#
# $1 - DICOM date
# $2 - DICOM time
# $3 - Timezone offset  
#
to_date_time_() {
	local da="$1"
	local tm="$2"
	local offset="$3"
	
	# sql formatted date
	local sql_d
	# sql formatted time
	local sql_t
	
	# sql formatted time
	local sql_datm
	# hhmmss part of time
	local tm_time=$( cut -d "." -f 1 <<< "$tm" )
	# fragments of seconds in time
	local tm_fragment=$( cut -d "." -f 2 -s <<< "$tm" )
	# time adjusted to 6 digits if was shorter
	local t6=$( sed 's;^\([[:digit:]]\{6\}\).*$;\1;' <<< "${tm_time}000000" )

	# convert date to sql format
	sql_d=$( sed -n 's;^\([[:digit:]]\{4\}\)\([[:digit:]]\{2\}\)\([[:digit:]]\{2\}\)$;\1-\2-\3;p' <<<"$da" )
	# check if date was valid
	[ -z "$sql_d" ] && return

	# convert hhmmss to sql format
	sql_t=$( sed -n 's;^\([[:digit:]]\{2\}\)\([[:digit:]]\{2\}\)\([[:digit:]]\{2\}\)\(\.[[:digit:]]+\)\?;\1:\2:\3\4;p' <<<"${t6}")

	if [ -z "$sql_t" ] ; then
		# invalid time
		sql_t="00:00:00"
	fi

	sql_datm=$(date -d "$sql_d $sql_t $offset" +"%F %T")
	[ $? -ne 0 ] && sql_datm=$(date -d "$sql_d $sql_t" +"%F %T")
	
	# append fragment if any
	[ -n "$sql_datm" -a -n "$tm_fragment" ] && sql_datm="${sql_datm}.$tm_fragment"

	echo "$sql_datm"
}


##
# Restores study in archive database.
#
# Checks if the study is obsoleted and
# restores archive data from PbR or image object
# with MIN SOP I UID
#
# Current directory is supposed to be expiration directory
# of the study.
#
#  $1 - Study Instance UID
# [$2 - force obsoleted files]
#
# force obsoleted files flag will only be set if all the files appears in
# the FILES list in obsoleted.rec of the study.
#
restore_study_()
{
	study="$1"
	[ -z "$study" ] && return 0
	USEOBSOLETED=""
	[ "$2" = "obsoleted" ] && USEOBSOLETED="true"

	hasReport=""
	hasImage=""

	MAINST=0
	TZOFFSET=""
	THEDATE=""
	THETIME=""
	PNAME=""
	PID=""
	PBDATE=""
	PBTIME=""
	ACCNO=""
	MODALITY=""
	BODYPART=""
	DESCRIPTION=""
	NUMOFOBJ=""
	ARCHST=10
	#SRST=1000

	# check and include obsoleted.rec
	WHOLE="false"
	FILES=""
	[ -f "$study/.info/obsoleted.rec" ] && . $study/.info/obsoleted.rec

	local archivedir=`pwd`
	archivedir=${archivedir#$ARCHIVE_DIR/}


	if [ "$notUseSQL" = "false"  ] && [ -f "$study/.info/insert.sql" ] ; then
		cat $study/.info/insert.sql | $MYSQL_HOME/bin/mysql -N $DBNAME 2>/dev/null
		# Update archive directory - for restored archive partition could be mounted to a different directory
		local updateClause="ARCHDIR='$archivedir'"
		if [ "$WHOLE" = "true" ] ; then
			# Update ARCHST in the database, it may be wrong due to a bug in obs.sh which does not update the
			# insert.sql query file
			updateClause="$updateClause,ARCHST=-100"
		fi
		$MYSQL_HOME/bin/mysql -N $DBNAME <<<"update $ARCHIVETBL set $updateClause where STYIUID='$study';" 2>/dev/null
	else

		# Set ARCHST to -100 if study is obsoleted
		[ "$WHOLE" = "true" ] && ARCHST=-100

		MODA=""   #Space separated distinct list of modalities
		hasPbR=""
		statusFileList=""
	  
		actObjList="$TMP/objList.$$.$study"
		rm -f "$actObjList" 2>/dev/null

		HAS_NON_OBSOLETED=""

		for object in `ls $study` ; do

			# check obsoleted files
			if [ -n "$FILES" ] ; then
				if is_obsoleted_ "$object" "$FILES" ; then
					[ -z "$USEOBSOLETED" ] && continue
				else
					HAS_NON_OBSOLETED="true"
				fi
			fi

			echo "$object" >>$actObjList

			# Skip PbD files
			[ "${object#PbD}" != "$object" ] && continue

			if [ "${object#PbR.}" != "$object" ] ; then
				#report
				if [ -z "$hasReport" ] ; then
					hasPbR="$object"
					hasReport="true"
				fi
				continue
			fi

			if [ "${object#SR.}" != "$object" -o  "${object#SRt.}" != "$object" ] ; then
				#Structured Report
				#SST="`$DCMTK_HOME/bin/dcmdump +P "f215,ba02" $study/$object | \
				#			sed 's/^.*\[\(.*\)\].*$/\1/'`"
				#if [ -n "$SST" -a "$SST" -lt $SRST ] ; then
				#	SRST=$SST
				#fi
				statusFileList="$statusFileList $object"
			else
				#image
				if [ -z "$hasImage" ] ; then
					hasImage="true"
				fi
				if [ "$quickMode" != "true" ] ; then
					MODY=`getDCMvalue "0008,0060" "$study/$object"`
					[ -z "$MODY" ] && continue
					newMody=`echo "$MODA " | grep " $MODY "`
					if [ -z "$newMody" ] ; then
						MODA="$MODA $MODY"
					fi
				fi
			fi
		done

		if [ -z "$HAS_NON_OBSOLETED" -a -z "$USEOBSOLETED" ] ; then
			# all the files in study are obsoleted and no force obsoleted flag was set
			restore_study_ "$study" obsoleted
			return $?
		fi

		#search MIN-SOPI-UID in precedence-order (modality,SCmodality,PbR,other...)
		precSorter=`cat "$actObjList" | sed 's/\./\t/1' | \
		awk '{
			prec = 10
			if ($1 == "SC")
				prec = 20
			else if ($1 == "PbR")
			  prec = 30
			else if ($1 == "PbD" || $1 == "SRt")
			  prec = 40
			print prec" "$2" "$1"."$2
		}'`
		sorted=`echo "$precSorter" | sort -k1,2`
		minSOPobj=`echo "$sorted" | grep -v "PbR" | head -n 1 | cut -f 3 -d " "`
		
		[ -n "$minSOPobj" ] && eval `$DCMTK_HOME/bin/dcmdump $study/$minSOPobj | $GETDATA_AWK`
		
		saveTZOFFSET="$TZOFFSET"
		saveTHEDATE="$THEDATE"
		saveTHETIME="$THETIME"
		NUMOFOBJ=`cat $actObjList | wc -l`
		rm -f "$actObjList"
		
		if [ -n "$hasPbR" ] ; then
			# SoftEdit correction
			eval `$DCMTK_HOME/bin/dcmdump $study/$hasPbR | $GETDATA_AWK`
			PbRstyDate=`getDCMvalue "0008,0020" "$study/$hasPbR"`
			if [ -z "$PbRstyDate" -a -n "$saveTHEDATE" ] ; then
				THEDATE="$saveTHEDATE"
				THETIME="$saveTHETIME"
				TZOFFSET="$saveTZOFFSET"
			fi
		fi

		# Concatenate modalities
		if [ -n "$MODA" ] ; then
			MODALITY=""
			for MODY in $MODA ; do
				[ -n "$MODALITY" ] && MODALITY="$MODALITY\\\\"
				MODALITY="$MODALITY$MODY"
			done
		else
			[ -n "$hasPbR" ] && MODALITY=`getDCMvalue "0008,0060" "$study/$hasPbR" | sed 's/\\\/\\\\\\\/g'`
		fi

		# MIN Search for Status
		[ -z "$statusFileList" ] && statusFileList="$hasPbR"
		if [ -n "$statusFileList" ] ; then
			MAINST=1000000
			for obj in $statusFileList ; do
				stat=`getDCMvalue "f215,1002" "$study/$obj"`
				[ -n "$stat" ] && [ $stat -lt $MAINST ] && MAINST=$stat
			done
			[ 1000000 -eq $MAINST ] && MAINST=0
		fi

		# do the database update
		FIELDS="MAINST,THEDATETIME,PNAME,PID,PBDATE,PBTIME,STYIUID,ACCNO,STYDESCR,MODALITY,BODYPART,NUMOFOBJ,ARCHDIR,ARCHST"

		THEDATETIME=$(to_date_time_ "$THEDATE" "$THETIME" "$TZOFFSET")

		if [ "$dumpOnly" != "true" ] ; then
			[ ! -d "$study/.info" ] && mkdir "$study/.info"

			tmpfile="$study/.info/.tmp.insert.sql.$$"
			echo "REPLACE INTO $ARCHIVETBL ($FIELDS) VALUES($MAINST,'$THEDATETIME','$PNAME','$PID','$PBDATE','$PBTIME','$study','$ACCNO','$DESCRIPTION','$MODALITY','$BODYPART',$NUMOFOBJ,'$archivedir',$ARCHST);" > "$tmpfile"
			$MYSQL_HOME/bin/mysql $DBNAME < "$tmpfile"

			if [ $? -ne 0 ] ; then
				echo "An error occured while replacing $study to the archive database" >&2
				rm "$tmpfile"
				return 1
			fi

			mv -f "$tmpfile" "$study/.info/insert.sql"
		else
			echo "REPLACE INTO $ARCHIVETBL ($FIELDS) VALUES($MAINST,'$THEDATETIME','$PNAME','$PID','$study','$ACCNO','$DESCRIPTION','$MODALITY','$BODYPART',$NUMOFOBJ,'$archivedir',$ARCHST);"
		fi
	fi

	return 0;
}

##
# Restores studies in day directory
#
# Current directory is supposed to be a day directory
#
restore_daydir_()
{
	for study in `ls` ; do

		# skip regular files
		[ ! -d "$study" ] && continue

		restore_study_ "$study"

	done # study
}


##
# Restores studies in expiration directory but if it contains day dir it calls the restor_daydir_
#
# Current directory is supposed to be an expiration directory
#
restore_expiredir_()
{
	for study in `ls` ; do

		# skip regular files
		[ ! -d "$study" ] && continue

		if [[ "$study" == [0-9][0-9] || "$study" == [0-9] ]] ; then
			# It's a 'day' directory with studies in it
			cd $study
			if [ $? -ne 0 ] ; then
				echo
				echo "Can not change current directory to"`pwd`"/$study : skipped" >&2
			else
				restore_daydir_
				RES=$?
				cd ..
			fi
		else
			restore_study_ "$study"
		fi

	done # study
}


##
# Iterates through archive directories
#
#  parameter
#    $1 : mode {check|mountlevel|[archivelevel]}
#
#  check:
#     Current directory is supposed to be $ARCHIVE_DIR.
#     Checks if mount points are mounted correctly
#     Iterates through current directory, skips regular files
#     and expiration directories. All other directories are
#     supposed to be mount points.
#     Tries to change current directory to mount points, to check
#     if the mount points are mounted correctly.
#
#  mountlevel:
#     Current directory is supposed to be $ARCHIVE_DIR/<mount point>.
#     Iterates through current directory and restores db
#     from the content of expiration directories.
#
#  archivelevel:
#     Current directory is supposed to be $ARCHIVE_DIR.
#     Iterates through current directory and restores db
#     from the content of expiration directories and mount points.
#
iterate_dirs_()
{
	CHECKMOUNT=""
	MOUNTLEVEL=""

	[ "$1" = "check" ] && CHECKMOUNT="true"
	[ "$1" = "mountlevel" ] && MOUNTLEVEL="true"

	CHECKRES=""

	for dir in `ls` ; do

		# skip regular files
		[ ! -d "$dir" ] && continue

		########################
		# expiration directories
		########################

		if [[ "$dir" == [0-9][0-9][0-9][0-9][0-9][0-9] || "$dir" == "unknown" ]] ; then
			# Skip them in CHECKMOUNT mode
			[ "$CHECKMOUNT" ] && continue

			cd $dir

			if [ $? -ne 0 ] ; then
				echo
				echo "Can not change current directory to $ARCHIVE_DIR/$dir : skipped" >&2
			else
				restore_expiredir_
				RES=$?
				cd ..
			fi
		fi

		##############
		# mount points
		##############

		[ -n "$MOUNTLEVEL" ] && continue

		cd $dir

		if [ $? -ne 0 ] ; then
			if [ -n "$CHECKMOUNT" ] ; then
				CHECKRES="$CHECKRES $dir"
			else
				echo "Can not change current directory to mount point $ARCHIVE_DIR/$dir : skipped" >&2
			fi
		else
			if [ -z "$CHECKMOUNT" ] ; then
				iterate_dirs_ mountlevel
			fi
			cd ..
		fi

	done

	MOUNTLEVEL=""

	[ -n "$CHECKMOUNT" ] && echo "$CHECKRES"
}


###########################################################
#                      MAIN
###########################################################


$MYSQL_HOME/bin/mysqlshow $DBNAME $ARCHIVETBL >/dev/null

if [ $? -ne 0 ] ; then
	echo
	echo "Archive table does not exist or mySQL is not running."
	echo "Please make sure that the mySQL server is started, the"
	echo "Archive=\"true\" flag is set in $ARCHIVE_VAR/archive.cfg and the"
	echo "Archive table exists in the $DBNAME database"
	exit 1
fi

if [ "$fixAll" = "true" ] ; then

	CWD=`pwd`
	cd $ARCHIVE_DIR

	if [ $? -ne 0 ] ; then
		echo
		echo "Can not change current directory to $ARCHIVE_DIR."
		echo "Please make sure that $ARCHIVE_DIR exists and is"
		echo "mounted correctly"
		exit 1
	fi

	# check mount points first

	CHECKRES=`iterate_dirs_ check`

	if [ -n "$CHECKRES" ] ; then
		echo
		echo "The following mount points in $ARCHIVE_DIR are not mounted correctly."
		echo "Please check and fix them before running restorearchivedb.sh."
		for i in $CHECKRES ; do
			echo "   $i"
		done
		exit 1
	fi

	OUTFILE="$TMP/restorearchive.$$.out"
	ERRFILE="$TMP/restorearchive.$$.err"

	echo
	echo "Start restoring archive database"
	echo "Standard output is redirected to $OUTFILE"
	echo "Standard error is redirected to $ERRFILE"


	iterate_dirs_ >$OUTFILE 2>$ERRFILE

	cd $CWD
elif [ -n "$studydir" ] ; then 
	
	echo "Fix study dir: $studydir"
	archdir=`dirname $studydir`
	study=`basename $studydir`
	echo "$archdir   $study" 

	cd $ARCHIVE_DIR/$archdir
	if [ $? -ne 0 ] ; then
		echo "Could not change directory to $ARCHIVE_DIR/$archdir" >&2
		exit 1
	fi
	
	restore_study_ "$study"
	res=$?

	cd $CWD

	exit $res
fi

echo "Done"
