#!/bin/bash
#
# Component: util
#
# Description:
#  Compresses a DICOM object.
#
# Usage:
#   compressObject.sh -i <input-object> [ -o <output-object> ] [ -v ]
#
#    -i input-object  :  DICOM object filename to be compressed
#    -o output-object :  filename for the compressed file (equals to input-object if not specified)
#    -v		      :  verbose operation.
#    -t tabulation    :  prefix to be used for writing strings out.
#
# Last Updated     : $Author$
# Update Date Time : $Date$
# RCS File Name    : $RCSfile$
# RCS Revision     : $Revision$
# Locked by        : $Locker$
# Symbolic name    : $Name$
# State            : $State$
#

. $HOME/etc/setenv.sh

CompressionMode="None"
MinLossyRatio="2.5"
MinLosslessRatio="1.5"
[ -f "$ARCHIVE_VAR/archive.cfg" ] && . $ARCHIVE_VAR/archive.cfg

LOGGER="/usr/bin/logger"
TAG="utils/compressObject[$$]"

ErrMsg=""

tabulation=""

# Parameters

Input=""

while getopts 'o:i:t:v' OPT; do
	case $OPT in
		i) Input="$OPTARG" ;;
		o) Output="$OPTARG" ;;
		t) tabulation="$OPTARG" ;;
		v) Verbose="true" ;;
		*) ErrMsg="Invalid_Parameter:$OPT" ;;
	esac
done

[ -z "$Output" ] && Output="$Input"

run_()
{
	if [ -z "$Input" ] ; then
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "missing parameter(s)"
		return 125
	fi

	if [ ! -f "$Input" ] ; then
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "Missing File $Input"
		return 125
	fi

	# Check if CompressionMode is set and compress DICOM file before putting to archive
	CompressedFile=""
	[ -n "$Verbose" ] && echo "${tabulation}$CompressionMode"
	[ "$CompressionMode" = "None" ] && return 0

	# Check if the DICOM file contains Pixel Data at all and if it's already compressed or not
	# dump TS and pixel data
 	ImageInfo=`$DCMTK_HOME/bin/dcmdump +P "0002,0010" +P "7fe0,0010" $Input 2>/dev/null`

	[ -n "$Verbose" ] && echo "$ImageInfo"

	if ! grep -q "7fe0,0010" <<<"$ImageInfo"; then
		[ -n "$Verbose" ] && echo "${tabulation}no image..."
		return 0
	fi

	if ! grep -q "0002,0010.*Endian" <<<"$ImageInfo"; then
		[ -n "$Verbose" ] && echo "${tabulation}Already Compressed/Encapsulated..."
		return 0
	fi


	CompressedTempFile="$TMP/compressObject.sh.$$"
	objdiffFlag=""
	if [ "$CompressionMode" = "JPEGLossless" -o "$CompressionMode" = "JPEGLosslessSV1" ] ; then
		CompFlag="+e1"
		[ "$CompressionMode" = "JPEGLossless" ] && CompFlag="+el"
		$DCMTK_HOME/bin/dcmcjpeg ${CompFlag} $Input ${CompressedTempFile} 2>/dev/null

		if [ $? -ne 0 ] ; then
			[ -f "${CompressedTempFile}" ] && rm "${CompressedTempFile}"
			[ -n "$Verbose" ] && echo "${tabulation}compression failed"
			return 1
		fi
	fi
	if [ "$CompressionMode" = "JPEG2000Lossless" -o "$CompressionMode" = "JPEG2000Lossy" ] ; then
		CompFlag="+tl"
		if [ "$CompressionMode" = "JPEG2000Lossy" ] ; then
			CompFlag="+ty"
			isLossy="true"
		fi

		$UTILS_HOME/bin/convert ${CompFlag} ${Input} ${CompressedTempFile} #2>/dev/null

		if [ $? -ne 0 ] ; then
			[ -f "${CompressedTempFile}" ] && rm "${CompressedTempFile}"
			[ -n "$Verbose" ] && echo "${tabulation}compression failed"
			return 1
		fi
	fi


	# Check if compressed file is equivalent to the original one
	$UTILS_HOME/bin/objdiff ${Input} ${CompressedTempFile} >/dev/null 2>&1
	if [ $? -ne 0 ] ; then
		[ -f "${CompressedTempFile}" ] && rm "${CompressedTempFile}"
		[ -n "$Verbose" ] && echo "${tabulation}objdiff failed (${CompressedTempFile})"
		return 1
	fi

	CompressedFile="$CompressedTempFile"


	# check if the compressed file is small enough
	origSize=`ls -l "$Input" | awk '{print $5}'`
	compSize=`ls -l "$CompressedFile" | awk '{print $5}'`
	ratio="$MinLossyRatio"
	[ -z $isLossy ] && ratio="$MinLosslessRatio"
	awk -v orig="$origSize" -v comp="$compSize" -v ratio="$ratio"  'BEGIN { if ( orig < comp*ratio) exit 1 } '
	if [ $? != 0 ] ; then
		# we don't use the compressed form, because we don't gain enough space 
		[ -n "$Verbose" ] && echo "${tabulation}compressed size($compSize) is greater than original size ($origSize) * ratio ($ratio)"
		[ -f "${CompressedFile}" ] && rm "${CompressedFile}"
		return 1
	fi
	if [ -n "${CompressedFile}" ] ; then 
		OutputTmp="${Output}.compressed.$$"
		cp -f $CompressedFile $OutputTmp
		cmp -s $CompressedFile $OutputTmp
		if [ $? != 0 ] ; then
			[ -n "$Verbose" ] && echo "${tabulation}Copying compressed object to target directory failed."
			[ -f "${CompressedFile}" ] && rm "${CompressedFile}"
			[ -f "${OutputTmp}" ] && rm "${OutputTmp}"
			return 1
		fi
		rm ${CompressedFile}
		[ -f "$Output" ] && mv "$Output" "$Output.orig.$$"
		mv "$OutputTmp" "$Output"
		if [ $? == 0 ] ; then
			rm "$Output.orig.$$" 2>/dev/null
		else
			[ -n "$Verbose" ] && echo "${tabulation}Renaming temp file to $Output failed."
			return 1
		fi
	fi
	return 0
}

# override print_
print_()
{
	echo "Compressing $Input"
}


get_watson_message_()
{
	if [ "$RUN_RESULT" -gt 124 ] ; then
		WATSON_MESSAGE_ID="compress.$Input.failed"
		WATSON_MESSAGE_TEXT="FAILED: compressing $Input"
		return
	fi

	if [ "$RUN_RESULT" -ne 0 -a "$RETRY_NUM" -gt 0 -a $(( $RETRY_NUM % 10)) -eq 0 ] ; then
		WATSON_MESSAGE_ID="compress.$Input.$RETRY_NUM"
		WATSON_MESSAGE_TEXT="ERROR: compressing $Input [Retry #$RETRY_NUM]"
	fi
}

run_
